# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.


from unittest.mock import Mock, patch

from pytest import fixture, mark, raises
from qiskit.primitives import BaseEstimator

from zne import ZNEStrategy
from zne.meta import zne
from zne.meta.cls import _get_zne_strategy, _set_zne_strategy

from .. import TYPES


class TestZNE:
    ################################################################################
    ## FIXTURES
    ################################################################################
    @fixture(scope="function")
    def Estimator(self):
        class Estimator(BaseEstimator):
            def _call(self):
                pass

            def _run(self):
                pass

        return Estimator

    @fixture(scope="function")
    def ZNEE(self, Estimator):
        return zne(Estimator)

    ################################################################################
    ## TESTS
    ################################################################################
    def test_zne_cls(self, Estimator):
        mocks = {p: Mock(return_value=p) for p in ("zne_init", "zne_call", "zne_run")}
        with patch.multiple("zne.meta.cls", **mocks):
            ZNEE = zne(Estimator)
        assert ZNEE.__init__ == "zne_init"
        mocks["zne_init"].assert_called_once()
        assert ZNEE.__call__ == "zne_call"
        mocks["zne_call"].assert_called_once()
        assert ZNEE._run == "zne_run"
        mocks["zne_run"].assert_called_once()
        assert hasattr(ZNEE, "zne_strategy")
        assert isinstance(ZNEE.zne_strategy, property)

    @mark.parametrize("obj", TYPES)
    def test_zne_cls_type_error(self, obj):
        with raises(TypeError):
            zne(obj)
        with raises(TypeError):
            zne(type(obj))

    def test_zne_strategy(self, ZNEE, Estimator):
        estimator = ZNEE()

        # None zne_strategy
        estimator.zne_strategy = None
        assert isinstance(estimator.zne_strategy, ZNEStrategy)
        assert estimator.zne_strategy is estimator._zne_strategy

        # Invalid zne_strategy
        with raises(TypeError):
            estimator.zne_strategy = "ERROR"

        # Custom zne_strategy
        zne_strategy = ZNEStrategy()
        estimator.zne_strategy = zne_strategy
        assert estimator.zne_strategy is zne_strategy
        assert estimator.zne_strategy is estimator._zne_strategy

        # Unset zne_strategy
        estimator = Estimator()  # instantiation before class update
        Estimator.zne_strategy = property(_get_zne_strategy, _set_zne_strategy)
        with raises(AttributeError):
            estimator._zne_strategy
        assert isinstance(estimator.zne_strategy, ZNEStrategy)
        assert estimator.zne_strategy is estimator._zne_strategy



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Serialization utils module."""

from __future__ import annotations

from datetime import datetime
from json import JSONEncoder, dump, dumps
from typing import Any

from numpy import ndarray
from qiskit.primitives import EstimatorResult


class DumpEncoder(JSONEncoder):
    """JSON encoder with extra methods for dumping."""

    @classmethod
    def dump(
        cls,
        obj: Any,
        file: str | None = None,
        indent: int | None = None,
    ) -> None:
        """Dump object to file using self as JSON encoder."""
        if file is None:  # pragma: no cover
            file = f"{datetime.utcnow().isoformat()}Z.json"
        with open(file, "w", encoding="utf8") as fp:  # pylint: disable=invalid-name
            return dump(obj, fp, indent=indent, cls=cls)

    @classmethod
    def dumps(cls, obj: Any, indent: int | None = None) -> str:
        """Dump object to string using self as JSON encoder."""
        return dumps(obj, indent=indent, cls=cls)


class ReprEncoder(DumpEncoder):
    """JSON encoder that falls back to `repr` if TypeError is raised."""

    def default(self, o):
        try:
            return super().default(o)
        except TypeError:
            return repr(o)


class NumPyEncoder(DumpEncoder):
    """JSON encoder for NumPy's :class:`numpy.ndarray` objects."""

    def default(self, o):
        if isinstance(o, ndarray):
            return o.tolist()
        return super().default(o)


class EstimatorResultEncoder(NumPyEncoder, ReprEncoder):
    """JSON encoder for :class:`EstimatorResult` objects."""

    def default(self, o):
        if isinstance(o, EstimatorResult):
            return {"values": o.values, "metadata": o.metadata}
        return super().default(o)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""ZNE strategy class to provide common base functionality."""

from __future__ import annotations

from inspect import signature
from typing import Any


class ImmutableStrategy:  # Note: pending deprecation  # pragma: no cover
    """Immutable strategy class to provide common base functionality.

    Note that all :class:`ImmutableStrategy` instance objects and their attributes are immutable by
    construction. If another strategy with different options needs to be utilized, a new instance
    of the class should be created.
    """

    def __new__(cls, *args, **kwargs) -> ImmutableStrategy:
        """ZNE strategy constructor."""
        self = super().__new__(cls)
        self.__dict__.update({"_init_options": {}})
        self._save_init_options(*args, **kwargs)
        return self

    def _save_init_options(self, *args, **kwargs) -> None:
        """Saves strategy init options as a dict."""
        init_signature = signature(self.__init__)  # type: ignore
        bound_signature = init_signature.bind(*args, **kwargs)
        bound_signature.apply_defaults()
        bound_args = bound_signature.arguments
        self._init_options: dict = dict(sorted(bound_args.items()))

    def __init__(self) -> None:
        """Base init method with no arguments, can be overriden."""

    @property
    def name(self) -> str:
        """Strategy name."""
        return type(self).__name__

    @property
    def options(self) -> dict:
        """Strategy options."""
        return self._init_options

    def __repr__(self) -> str:
        return f"<{self.name}:{self.options}>" if self.options else self.name

    def __eq__(self, other: Any) -> bool:
        return repr(self) == repr(other)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from inspect import signature
from random import seed as set_random_seed
from random import shuffle
from unittest.mock import Mock

from pytest import fixture, mark

from zne import NOISE_AMPLIFIER_LIBRARY
from zne.immutable_strategy import ImmutableStrategy

STRATEGIES = {**NOISE_AMPLIFIER_LIBRARY}

INIT_OPTIONS = [
    ((), {}),
    ((), {"hello": "world"}),
    ((1, 2), {}),
    ((1.1, 2.2), {"Qiskit": True}),
    (("A", 12), {"pain": None, "gain": None}),
]


class TestImmutableStrategy:

    ################################################################################
    ## FIXTURES
    ################################################################################
    @fixture
    def DummyStrategy(self):
        class DummyStrategy(ImmutableStrategy):
            _save_init_options = Mock()

            def __init__(self, *args, **kwargs):
                pass

        return DummyStrategy

    @fixture
    def get_init_options(self):
        def factory_method(cls):
            init_signature = signature(cls.__init__)
            init_keys = init_signature.parameters.keys()
            init_options = {k: f"Mock({k})" for k in init_keys if k != "self"}
            return init_options

        return factory_method

    @fixture
    def shuffle_init_options(self, get_init_options):
        def factory_method(cls, seed=None):
            set_random_seed(seed)
            ordered_options = get_init_options(cls)
            items = list(ordered_options.items())
            shuffle(items)
            return dict(items)

        return factory_method

    ################################################################################
    ## TESTS
    ################################################################################
    @mark.parametrize(
        "init_args, init_kwargs",
        [
            ((), {}),
            ((), {"hello": "world"}),
            ((1, 2), {}),
            ((1.1, 2.2), {"Qiskit": True}),
            (("A", 12), {"pain": None, "gain": None}),
        ],
    )
    def test_new(self, init_args, init_kwargs, DummyStrategy):
        strategy = DummyStrategy(*init_args, **init_kwargs)
        strategy._save_init_options.assert_called_once_with(*init_args, **init_kwargs)

    @mark.parametrize("init_args, init_kwargs", INIT_OPTIONS)
    def test_save_init_options(self, init_args, init_kwargs, DummyStrategy):
        strategy = DummyStrategy()
        assert strategy._init_options == {}
        strategy._save_init_options(*init_args, **init_kwargs)
        assert strategy._init_options == {}

    @mark.parametrize("Strategy", STRATEGIES.values(), ids=STRATEGIES.keys())
    def test_name(self, Strategy):
        strategy = Strategy()
        assert strategy.name == Strategy.__name__

    @mark.parametrize("Strategy", STRATEGIES.values(), ids=STRATEGIES.keys())
    def test_options(self, Strategy, shuffle_init_options):
        options = shuffle_init_options(Strategy, seed=0)
        strategy = Strategy.__new__(Strategy, **options)
        options.pop("warn_user", None)
        assert strategy.options == dict(sorted(options.items()))

    @mark.parametrize("Strategy", STRATEGIES.values(), ids=STRATEGIES.keys())
    def test_repr(self, Strategy):
        strategy = Strategy()
        options = dict({"example": "option"})
        strategy._init_options = options
        assert repr(strategy) == f"<{Strategy.__name__}:{options}>"

    @mark.parametrize("Strategy", STRATEGIES.values(), ids=STRATEGIES.keys())
    def test_eq(self, Strategy, shuffle_init_options):
        options_1 = shuffle_init_options(Strategy, seed=0)
        options_2 = shuffle_init_options(Strategy, seed=1)
        assert Strategy.__new__(Strategy, **options_1) == Strategy.__new__(Strategy, **options_2)



#!/usr/bin/env python3

# This code is part of Qiskit.
#
# (C) Copyright IBM 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import re
import configparser

import toml
import fire


def mapfunc_dev(dep):
    """Load the development version(s) of certain Qiskit-related packages"""
    # https://peps.python.org/pep-0440/#direct-references
    return re.sub(
        r"^(qiskit).*$",
        r"\1 @ git+https://github.com/Qiskit/\1.git",
        dep,
    )


def mapfunc_min(dep):
    """Set each dependency to its minimum version"""
    return re.sub(r"[>~]=", r"==", dep)


def inplace_map(fun, lst: list):
    """In-place version of Python's `map` function"""
    for i, x in enumerate(lst):
        lst[i] = fun(x)


def process_dependencies_in_place(d: dict, mapfunc):
    """Given a parsed `pyproject.toml`, process dependencies according to `mapfunc`"""
    proj = d["project"]

    try:
        deps = proj["dependencies"]
    except KeyError:
        pass  # no dependencies; that's unusual, but fine.
    else:
        inplace_map(mapfunc, deps)

    try:
        opt_deps = proj["optional-dependencies"]
    except KeyError:
        pass  # no optional dependencies; that's fine.
    else:
        for dependencies_list in opt_deps.values():
            inplace_map(mapfunc, dependencies_list)

    try:
        build_system = d["build-system"]
    except KeyError:
        pass
    else:
        try:
            build_system_requires = build_system["requires"]
        except KeyError:
            pass
        else:
            inplace_map(mapfunc, build_system_requires)


class CLI:
    """Command-line interface class for Fire"""

    def get_tox_minversion(self):
        """Extract tox minversion from `tox.ini`"""
        config = configparser.ConfigParser()
        config.read("tox.ini")
        print(config["tox"]["minversion"])

    def pin_dependencies(self, strategy, inplace: bool = False):
        """Pin the dependencies in `pyproject.toml` according to `strategy`"""
        mapfunc = {
            "dev": mapfunc_dev,
            "min": mapfunc_min,
        }[strategy]

        with open("pyproject.toml") as f:
            d = toml.load(f)
        process_dependencies_in_place(d, mapfunc)

        if inplace:
            with open("pyproject.toml", "w") as f:
                toml.dump(d, f)
        else:
            print(toml.dumps(d))


if __name__ == "__main__":
    fire.Fire(CLI)



#!/usr/bin/env python3

##    _____  _____
##   |  __ \|  __ \    AUTHOR: Pedro Rivero
##   | |__) | |__) |   ---------------------------------
##   |  ___/|  _  /    DATE: Feb 21, 2023
##   | |    | | \ \    ---------------------------------
##   |_|    |_|  \_\   https://github.com/pedrorrivero
##

## Copyright 2023 Pedro Rivero
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
## http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

"""Create symlinks to submodule files in base repo, adding all missing directories."""

from pathlib import Path


EXCLUDE = {"README.md", ".git", ".gitignore"}


def symlink_files(original: Path, host_dir: Path, relative: Path = None) -> None:
    """Recursively create symlinks(s) to file(s) in original path into host directory.
    
    Args:
        original: path to original file or directory.
        host_dir: directory to host the symlinks.
        relative: relative path to build the symlinks from.
    """
    # Input standardization
    original = original.resolve()
    host_dir = host_dir.resolve()
    if not host_dir.is_dir():
        host_dir = host_dir.parent
    if relative is None:
        relative = Path()
    else:
        relative = relative.resolve()
    
    # Logic
    host = host_dir.joinpath(original.name)
    if original.is_file():
        original = original.relative_to(relative)
        host.symlink_to(original)
    elif original.is_dir():
        host.mkdir(parents=True, exist_ok=True)
        for path in original.iterdir():
            symlink_files(path, host, relative)


if __name__ == "__main__":
    file = Path(__file__).resolve()
    tools = file.parent
    submodule = tools.parent
    repo = submodule.parent
    for file in submodule.iterdir():
        if file.name not in EXCLUDE:
            symlink_files(file, repo, repo)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Build dictionary library util."""

from qiskit.circuit import ControlledGate, Gate, QuantumCircuit


def build_method_to_gate_dict() -> dict:
    """Returns dictionary mapping gate names to gate classes."""
    method_to_gate = {}
    gates = Gate.__subclasses__() + ControlledGate.__subclasses__()
    for gate in gates:
        name = gate.__name__.lower()
        if name[-4:] == "gate":
            method = name[:-4]
            if hasattr(QuantumCircuit, method):
                method_to_gate[method] = gate
    return method_to_gate


STANDARD_GATES = frozenset(build_method_to_gate_dict())



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from unittest.mock import Mock

from numpy import array
from pytest import fixture, mark, raises

from zne.extrapolation import Extrapolator, OLSExtrapolator
from zne.extrapolation.extrapolator import ReckoningResult
from zne.utils.typing import normalize_array
from zne.utils.unset import UNSET


class TestExtrapoaltor:
    """Test Extrapolator interface."""

    ################################################################################
    ## FIXTURES
    ################################################################################
    @fixture(scope="function")
    def MockExtrapolator(self):
        class MockExtrapolator(Extrapolator):
            def __init__(self, min_points=2) -> None:
                self._min_points = min_points

            @property
            def min_points(self):
                return self._min_points

            def _extrapolate_zero(self, x_data, y_data, sigma_x, sigma_y):
                return ReckoningResult(1, 1, {})

        return MockExtrapolator

    @fixture(scope="function")
    def extrapolator(self, MockExtrapolator):
        return MockExtrapolator()

    ################################################################################
    ## TESTS
    ################################################################################
    @mark.parametrize(
        "x_data, y_data, sigma_x, sigma_y",
        [
            ([1, 2], [1, 2], None, None),
            ([1, 2], [1, 2], None, [4, 4]),
            ([1, 2], [1, 2], [4, 4], None),
            ([1, 2], [1, 2], [4, 4], [4, 4]),
            ([1, 2, 3], [1, 2, 3], None, None),
            ([1, 2, 3], [1, 2, 3], None, [4, 4, 4]),
            ([1, 2, 3], [1, 2, 3], [4, 4, 4], None),
            ([1, 2, 3], [1, 2, 3], [4, 4, 4], [4, 4, 4]),
        ],
    )
    def test_extrapolate_zero(self, extrapolator, x_data, y_data, sigma_x, sigma_y):
        """Test extrapolate zero."""
        extrapolator._extrapolate_zero = Mock()
        extrapolator.extrapolate_zero(x_data, y_data, sigma_x, sigma_y)
        extrapolator._extrapolate_zero.assert_called_once()
        call_args = extrapolator._extrapolate_zero.call_args[0]
        expected_call_args = (
            tuple(x_data),
            tuple(y_data),
            tuple(sigma_x or [1] * len(x_data)),
            tuple(sigma_y or [1] * len(y_data)),
        )
        for arg, exp in zip(call_args, expected_call_args):
            assert arg == exp

    @mark.parametrize(
        "x_data, y_data, sigma_x, sigma_y",
        [
            ([1, 2], [0, 1, 2], None, None),
            ([0, 1, 2], [1, 2], None, None),
            ([1, 2], [1, 2], [1, 1, 1], None),
            ([1, 2], [1, 2], None, [1, 1, 1]),
        ],
    )
    def test_extrapolate_zero_input_sizes(self, extrapolator, x_data, y_data, sigma_x, sigma_y):
        """Test extrapolate zero input sizes."""
        with raises(ValueError):
            extrapolator.extrapolate_zero(x_data, y_data, sigma_x, sigma_y)

    @mark.parametrize("data", [(1,), (1, 1), (1, 2)])
    def test_validate_min_points(self, MockExtrapolator, data):
        """Test validate data min points."""
        extrapolator = MockExtrapolator(len(set(data)) + 1)
        with raises(ValueError):
            extrapolator.extrapolate_zero(data, data)

    @mark.parametrize("data", [[0] * 2, [0] * 3, [0] * 4, [1] * 2, [1] * 3, [1] * 4, range(4)])
    def test_validate_data(self, extrapolator, data):
        """Test validate data."""
        result = extrapolator._validate_data(data)
        assert result == tuple(data)

    @mark.parametrize("data", [None, True, 1, 1.0, float("nan"), 1j, [1j], [None], set(), [[0]]])
    def test_validate_data_type_error(self, extrapolator, data):
        """Test validate data type error."""
        with raises(TypeError):
            extrapolator._validate_data(data)

    @mark.parametrize(
        "sigma", [[0] * 2, [0] * 3, [0] * 4, [1] * 2, [1] * 3, [1] * 4, range(4), None]
    )
    def test_validate_sigma(self, extrapolator, sigma):
        """Test validate sigma."""
        default_size = 4
        result = extrapolator._validate_sigma(sigma, default_size)
        if sigma is None:
            sigma = [1] * default_size
        assert result == tuple(sigma)


class TestOLSExtrapolator:
    """Test Extrapolator interface."""

    ################################################################################
    ## FIXTURES
    ################################################################################
    @fixture(scope="function")
    def MockExtrapolator(self):
        class MockExtrapolator(OLSExtrapolator):
            def __init__(self, min_points=2) -> None:
                self._min_points = min_points

            @property
            def min_points(self):
                return self._min_points

            def _extrapolate_zero(self, x_data, y_data, sigma_x, sigma_y):
                return ReckoningResult(1, 1, {})

            def _model(self, x, *coefficients):
                return 1

        return MockExtrapolator

    @fixture(scope="function")
    def extrapolator(self, MockExtrapolator):
        return MockExtrapolator()

    ################################################################################
    ## TESTS
    ################################################################################
    @mark.parametrize(
        "y_data, residuals, expected",
        [
            ([0, 2], [0, 0], 1),
            ([0, 2], [1, 0], 1 / 2),
            ([0, 2], [0, 1], 1 / 2),
            ([0, 2], [1, 1], 0),
            ([0, 1, 2], [0, 0, 0], 1),
            ([0, 1, 2], [1, 0, 0], 1 / 2),
            ([0, 1, 2], [0, 1, 0], 1 / 2),
            ([0, 1, 2], [0, 0, 1], 1 / 2),
            ([0, 1, 2], [1, 1, 0], 0),
            ([0, 1, 2], [1, 0, 1], 0),
            ([0, 1, 2], [0, 1, 1], 0),
            ([0, 1, 2], [1, 1, 1], -1 / 2),
        ],
    )
    def test_r_squared(self, extrapolator, y_data, residuals, expected):
        """Test R-squared."""
        y_data, residuals = array(y_data), array(residuals)
        assert extrapolator._r_squared(y_data, residuals) == expected

    @mark.parametrize(
        "x_data, y_data, coefficients, covariance_matrix, r_squared",
        [
            ([0, 1], [0, 1], [0, 1], [[0, 0], [0, 0]], 1),
            ([0, 1, 2], [0, 1, 2], [1, 0], [[1, 1], [1, 1]], 1),
        ],
    )
    def test_build_metadata(
        self, extrapolator, x_data, y_data, coefficients, covariance_matrix, r_squared
    ):
        """Test build metadata."""
        x_data, y_data = array(x_data), array(y_data)
        coefficients, covariance_matrix = array(coefficients), array(covariance_matrix)
        extrapolator._r_squared = Mock(return_value=r_squared)
        metadata = extrapolator._build_metadata(x_data, y_data, coefficients, covariance_matrix)
        for entry in ("coefficients", "covariance_matrix", "residuals", "R2"):
            assert metadata.get(entry, UNSET) is not UNSET
        residuals = y_data - extrapolator._model(x_data, *coefficients)
        assert metadata.get("coefficients") == normalize_array(coefficients)
        assert metadata.get("covariance_matrix") == normalize_array(covariance_matrix)
        assert metadata.get("residuals") == normalize_array(residuals)
        assert metadata.get("R2") == normalize_array(r_squared)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from unittest.mock import Mock

from pytest import fixture, mark, raises

from zne.meta.job import ZNEJob
from zne.meta.run import zne_run
from zne.zne_strategy import ZNEStrategy


class TestZNERun:
    ################################################################################
    ## FIXTURES
    ################################################################################
    @fixture(scope="function")
    def self(_self):
        self = Mock()
        self.zne_strategy = ZNEStrategy()
        return self

    @fixture(scope="function")
    def run_options(_self):
        return {"Forrest": "Gump", "Charriots": "Fire"}

    ################################################################################
    ## TESTS
    ################################################################################
    def test_base_run_noop_zne_strategy(_self, self, run_options):
        base_run = Mock()
        run = zne_run(base_run)
        self.zne_strategy = ZNEStrategy(noise_factors=range(1, 4))
        _ = run(
            self,
            circuits="circuits",
            observables="observables",
            parameter_values="parameter_values",
            zne_strategy=None,
            **run_options,
        )
        base_run.assert_called_once_with(
            self,
            circuits="circuits",
            observables="observables",
            parameter_values="parameter_values",
            **run_options,
        )

    def test_base_run_default_zne_strategy(_self, self, run_options):
        base_run = Mock()
        run = zne_run(base_run)
        self.zne_strategy = ZNEStrategy(noise_factors=range(1, 4))
        build_noisy_circuits = Mock(return_value="noisy_circuits")
        self.zne_strategy.build_noisy_circuits = build_noisy_circuits
        map_to_noisy_circuits = Mock(side_effect=lambda x: "noisy_" + x)
        self.zne_strategy.map_to_noisy_circuits = map_to_noisy_circuits
        _ = run(
            self,
            circuits="circuits",
            observables="observables",
            parameter_values="parameter_values",
            **run_options,
        )
        build_noisy_circuits.assert_called_once_with("circuits")
        map_to_noisy_circuits.assert_any_call("observables")
        map_to_noisy_circuits.assert_any_call("parameter_values")
        assert map_to_noisy_circuits.call_count == 2
        base_run.assert_called_once_with(
            self,
            circuits="noisy_circuits",
            observables="noisy_observables",
            parameter_values="noisy_parameter_values",
            **run_options,
        )

    def test_base_run_custom_zne_strategy(_self, self, run_options):
        base_run = Mock()
        run = zne_run(base_run)
        zne_strategy = ZNEStrategy(noise_factors=range(1, 7))
        build_noisy_circuits = Mock(return_value="noisy_circuits")
        zne_strategy.build_noisy_circuits = build_noisy_circuits
        map_to_noisy_circuits = Mock(side_effect=lambda x: "noisy_" + x)
        zne_strategy.map_to_noisy_circuits = map_to_noisy_circuits
        _ = run(
            self,
            circuits="circuits",
            observables="observables",
            parameter_values="parameter_values",
            zne_strategy=zne_strategy,
            **run_options,
        )
        build_noisy_circuits.assert_called_once_with("circuits")
        map_to_noisy_circuits.assert_any_call("observables")
        map_to_noisy_circuits.assert_any_call("parameter_values")
        assert map_to_noisy_circuits.call_count == 2
        base_run.assert_called_once_with(
            self,
            circuits="noisy_circuits",
            observables="noisy_observables",
            parameter_values="noisy_parameter_values",
            **run_options,
        )

    def test_zne_strategy_type_error(_self, self, run_options):
        base_run = Mock()
        run = zne_run(base_run)
        with raises(TypeError):
            run(
                self,
                circuits="circuits",
                observables="observables",
                parameter_values="parameter_values",
                zne_strategy="zne_strategy",
                **run_options,
            )

    def test_zne_job(_self, self, run_options):
        base_run = Mock(return_value="job")
        run = zne_run(base_run)
        job = run(
            self,
            circuits="circuits",
            observables="observables",
            parameter_values="parameter_values",
            **run_options,
        )
        assert isinstance(job, ZNEJob)
        assert job.base_job == "job"
        assert job.zne_strategy is self.zne_strategy
        assert job.target_num_experiments == len("circuits")


@mark.parametrize("run", [None, Ellipsis, True, 0, 1.0, 1j, "run"])
def test_type_error(run):
    with raises(TypeError):
        zne_run(run)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Type checking utils module."""

from typing import Any

from numpy import array


def isint(obj: Any) -> bool:
    """Check if object is int"""
    return isinstance(obj, int) and not isinstance(obj, bool)


def isinteger(obj: Any) -> bool:
    """Check if object is an integer number"""
    return isint(obj) or isinstance(obj, float) and obj.is_integer()


def isreal(obj: Any) -> bool:
    """Check if object is a real number: int or float minus ``±Inf`` and ``NaN``."""
    return isint(obj) or isinstance(obj, float) and float("-Inf") < obj < float("Inf")


def normalize_array(arr: Any) -> Any:
    """Normalize array-like objects (i.e. as in numpy) to tuples of the same shape.

    If zero dimensional, the object is preserved or casted to a core python type.
    """
    arr = array(arr)
    num_dimensions = len(arr.shape)
    if num_dimensions == 0:
        return arr.tolist()
    if num_dimensions == 1:
        return tuple(arr.tolist())
    return tuple(map(normalize_array, arr))



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Zero Noise Extrapolation (ZNE) strategy configuration dataclass."""

from __future__ import annotations

from collections.abc import Iterable, Iterator, Sequence
from math import sqrt
from typing import Any
from warnings import warn

from numpy import array
from qiskit import QuantumCircuit
from qiskit.primitives import EstimatorResult

from .extrapolation import Extrapolator, LinearExtrapolator
from .noise_amplification import MultiQubitAmplifier, NoiseAmplifier
from .types import EstimatorResultData, Metadata  # noqa: F401
from .utils.grouping import from_common_key, group_elements_gen, merge_dicts
from .utils.typing import isreal, normalize_array
from .utils.validation import quality


class ZNEStrategy:
    """Zero Noise Extrapolation strategy.

    Args:
        noise_factors: An list of real valued noise factors that determine by what amount the
            circuits' noise is amplified.
        noise_amplifier: A noise amplification strategy implementing the :class:`NoiseAmplifier`
            interface. A dictionary of currently available options can be imported as
            ``NOISE_AMPLIFIER_LIBRARY``.
        extrapolator: An extrapolation strategy implementing the :class:`Extrapolator`
            interface. A dictionary of currently available options can be imported as
            ``EXTRAPOLATOR_LIBRARY``.

    Raises:
        TypeError: If input does not match the type specification.
        ValueError: If noise factors are empty or less than one.
    """

    _DEFINING_ATTRS = (
        "noise_factors",
        "noise_amplifier",
        "extrapolator",
    )

    def __init__(
        self,
        noise_factors: Sequence[float] | None = None,
        noise_amplifier: NoiseAmplifier | None = None,
        extrapolator: Extrapolator | None = None,
    ) -> None:
        self.noise_factors = noise_factors
        self.noise_amplifier = noise_amplifier
        self.extrapolator = extrapolator

    def __repr__(self) -> str:
        attrs_str = ", ".join(
            f"{attr}={repr(getattr(self, attr))}" for attr in self._DEFINING_ATTRS
        )
        return f"ZNEStrategy({attrs_str})"

    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, self.__class__):
            return False
        return all(getattr(self, attr) == getattr(__o, attr) for attr in self._DEFINING_ATTRS)

    def __bool__(self) -> bool:
        return not self.is_noop

    ################################################################################
    ## CONSTRUCTORS
    ################################################################################
    @classmethod
    def noop(cls) -> ZNEStrategy:
        """Construct a no-op ZNE strategy object."""
        return cls(noise_factors=(1,))

    ################################################################################
    ## PROPERTIES
    ################################################################################
    @quality(default=(1,))
    def noise_factors(self, noise_factors: Sequence[float]) -> tuple[float, ...]:
        """Noise factors for ZNE.

        Validation logic defined as required by `quality`.
        """
        if not isinstance(noise_factors, Sequence):
            raise TypeError(
                f"Expected `Sequence` noise factors, received `{type(noise_factors)}` instead."
            )
        noise_factors = tuple(noise_factors)
        if not noise_factors:
            raise ValueError("Noise factors must not be empty.")
        if not all(isreal(nf) for nf in noise_factors):
            raise TypeError("Noise factors must be real valued.")
        if any(nf < 1 for nf in noise_factors):
            raise ValueError("Noise factors must be greater than or equal to one.")

        unsorted_noise_factors = noise_factors
        noise_factors = tuple(sorted(noise_factors))
        if noise_factors != unsorted_noise_factors:
            warn("Unordered noise factors detected and rearranged.", UserWarning, stacklevel=3)

        duplicate_noise_factors = noise_factors
        noise_factors = tuple(sorted(set(noise_factors)))
        if noise_factors != duplicate_noise_factors:
            warn("Duplicate noise factors detected and erased.", UserWarning, stacklevel=3)

        return noise_factors

    @quality(default=MultiQubitAmplifier())
    def noise_amplifier(self, noise_amplifier: NoiseAmplifier) -> NoiseAmplifier:
        """Noise amplifier strategy for ZNE.

        Validation logic defined as required by `quality`.
        """
        if not isinstance(noise_amplifier, NoiseAmplifier):
            raise TypeError(
                f"Expected `NoiseAmplifier` object, received `{type(noise_amplifier)}` instead."
            )
        return noise_amplifier

    @quality(default=LinearExtrapolator())
    def extrapolator(self, extrapolator: Extrapolator) -> Extrapolator:
        """Extrapoaltor strategy for ZNE.

        Validation logic defined as required by `quality`.
        """
        if not isinstance(extrapolator, Extrapolator):
            raise TypeError(
                f"Expected `Extrapolator` object, received `{type(extrapolator)}` instead."
            )
        return extrapolator

    @property
    def num_noise_factors(self) -> int:
        """Number of noise factors."""
        return len(self.noise_factors)

    @property
    def performs_noise_amplification(self) -> bool:
        """Checks if noise amplification is performed."""
        return any(nf > 1 for nf in self.noise_factors)

    @property
    def performs_zne(self) -> bool:
        """Checks if zero noise extrapolation is performed."""
        return self.performs_noise_amplification and self.num_noise_factors > 1

    @property
    def is_noop(self) -> bool:
        """Checks if strategy is no-op."""
        return not self.performs_noise_amplification and not self.performs_zne

    ################################################################################
    ## NOISE AMPLIFICATION
    ################################################################################
    def amplify_circuit_noise(self, circuit: QuantumCircuit, noise_factor: float) -> QuantumCircuit:
        """Noise amplification from :class:`~.noise_amplification.NoiseAmplifier`.

        Args:
            circuit: The original quantum circuit.
            noise_factor: The noise amplification factor by which to amplify the circuit noise.

        Returns:
            The noise amplified quantum circuit
        """
        # TODO: caching
        return self.noise_amplifier.amplify_circuit_noise(circuit, noise_factor)

    # TODO: decouple indexing logic depending on this method
    # TODO: add validation
    def build_noisy_circuits(
        self,
        original_circuits: Iterable[QuantumCircuit] | QuantumCircuit,
    ) -> tuple[QuantumCircuit, ...]:
        """Construct noisy circuits for all noise factors from original circuits.

        Args:
            original_circuits: a :class:`~qiskit.circuit.QuantumCircuit` or a collection of
                :class:`~qiskit.circuit.QuantumCircuit`.

        Returns:
            A tuple containing the noise amplified circuits.
        """

        def generate_noisy_circuits(
            original_circuits: Iterable[QuantumCircuit],
        ) -> Iterator[QuantumCircuit]:
            for circuit in original_circuits:  # type: QuantumCircuit
                for noise_factor in self.noise_factors:  # type: float
                    yield self.amplify_circuit_noise(circuit, noise_factor)

        if isinstance(original_circuits, QuantumCircuit):
            original_circuits = [original_circuits]
        return tuple(generate_noisy_circuits(original_circuits))

    # TODO: decouple indexing logic depending on this method
    def map_to_noisy_circuits(self, arg: Any) -> tuple | None:
        """Map arguments for original circuits to the corresponding arguments for noisy circuits.

        Args:
            arg: Additional non-circuit arguments such as observables or parameter values.

        Returns:
            A tuple of args corresponding to the noise amplified circuits or None.
        """
        if arg is None:
            return arg
        if not isinstance(arg, Iterable):
            arg = [arg]
        mapped_arg: list = []
        for element in arg:
            mapped_arg.extend(element for _ in range(self.num_noise_factors))
        return tuple(mapped_arg)

    ################################################################################
    ## EXTRAPOLATION
    ################################################################################
    # TODO: add validation
    def mitigate_noisy_result(self, noisy_result: EstimatorResult) -> EstimatorResult:
        """Parse results from noisy circuits to error mitigated results.

        Args:
            noisy_result: The unmitigated results.

        Returns:
            The mitigated results after performing zero-noise-extrapolation.
        """
        values: list[float] = []
        metadata: list[Metadata] = []
        for result_group in self._generate_noisy_result_groups(noisy_result):
            data = self._regression_data_from_result_group(result_group)
            val, err, meta = self.extrapolator.extrapolate_zero(*data)
            common_metadata: Metadata = {"std_error": err}  # TODO: extract other common metadata
            zne_metadata: Metadata = self.build_zne_metadata(result_group, meta)
            values.append(val)
            metadata.append({**common_metadata, "zne": zne_metadata})
        return EstimatorResult(values=array(values), metadata=list(metadata))

    # TODO: decouple indexing logic depending on this method
    def _generate_noisy_result_groups(
        self, noisy_result: EstimatorResult
    ) -> Iterator[EstimatorResult]:
        """Generator function for grouping noisy results.

        Iteratively constructs an estimator result for each group of experiments containing
        the measurement results associated with every noise factor.

        Args:
            noisy_result: The estimator result with data from all the experiments performed.

        Yields:
            Estimator results grouping data for experiments with different noise factors,
            but same circuit-observable combinations.

        Raises:
            ValueError: If the number of performed experiments is not an integer mutliple of
                the number of noise factors.
        """
        if noisy_result.num_experiments % self.num_noise_factors != 0:
            raise ValueError("Inconsistent number of noisy experiments and noise factors.")
        for group in group_elements_gen(
            noisy_result.experiments, group_size=self.num_noise_factors
        ):  # type: tuple[EstimatorResultData, ...]
            values, metadata = zip(*[data.values() for data in group])
            yield EstimatorResult(values=array(values), metadata=list(metadata))

    def _regression_data_from_result_group(
        self, result_group: EstimatorResult
    ) -> tuple[list[float], list[float], list[float], list[float]]:
        """Build regression data from noisy result group.

        Args:
            result_group: Estimator result grouping data for experiments with different
                noise factors, but same circuit-observable combinations.

        Returns:
            Regression data
        """
        if result_group.num_experiments != self.num_noise_factors:
            raise ValueError("Inconsistent number of noisy experiments and noise factors.")
        x_data = list(self.noise_factors)  # TODO: get actual noise factors achieved
        y_data = result_group.values.tolist()
        sigma_x = [1 for _ in x_data]
        sigma_y = [sqrt(md.get("variance", 1)) for md in result_group.metadata]
        return x_data, y_data, sigma_x, sigma_y  # type: ignore

    # TODO: add validation
    def build_zne_metadata(
        self, result_group: EstimatorResult, extrapolation: Metadata | None = None
    ) -> Metadata:
        """Build ZNE metadata from extrapolation data.

        Args:
            result_group: The grouped noisy results for a single mitigated result.
            extrapolation: extrapolation metadata entries to include (e.g. extrapolation).

        Returns:
            Dictionary containing ZNE metadata.

        Raises:
            ValueError: If the number of experiments does not match the number of noise factors.
        """
        if extrapolation is None:
            extrapolation = {}
        if result_group.num_experiments != self.num_noise_factors:
            raise ValueError("Inconsistent number of noisy experiments and noise factors.")
        noise_amplification: Metadata = {
            "noise_amplifier": self.noise_amplifier,
            "noise_factors": self.noise_factors,
            "values": normalize_array(result_group.values),  # TODO: simplify when tuple
        }
        for key in merge_dicts(result_group.metadata):
            value = from_common_key(result_group.metadata, key)
            noise_amplification.update({key: value})
        extrapolation = {
            "extrapolator": self.extrapolator,
            **extrapolation,
        }
        return {"noise_amplification": noise_amplification, "extrapolation": extrapolation}



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from test import NO_INTS, NO_REAL

from numpy import array
from pytest import mark

from zne.utils.typing import isint, isinteger, isreal, normalize_array


################################################################################
## TYPE CHECKING
################################################################################
@mark.parametrize("object", [0, 1, -1])
def test_isint_true(object):
    """Test isint true."""
    assert isint(object)


@mark.parametrize("object", NO_INTS, ids=[str(type(i).__name__) for i in NO_INTS])
def test_isint_false(object):
    """Test isint false."""
    assert not isint(object)


@mark.parametrize("object", [0, 1, -1, 1.0, -1.0])
def test_isinteger_true(object):
    """Test isinteger true."""
    assert isinteger(object)


@mark.parametrize("object", [1.2, -2.4])
def test_isinteger_false(object):
    """Test isinteger false."""
    assert not isinteger(object)


@mark.parametrize("object", [0, 1, -1, 1.2, -2.4])
def test_isreal_true(object):
    """Test isreal true."""
    assert isreal(object)


@mark.parametrize("object", NO_REAL, ids=[str(type(i).__name__) for i in NO_REAL])
def test_isreal_false(object):
    """Test isreal false."""
    assert not isreal(object)


@mark.parametrize(
    "arr, expected",
    [
        (1, 1),
        (1.0, 1.0),
        (1j, 1j),
        (None, None),
        (dict(), dict()),
        (set(), set()),
        (list(), tuple()),
        (tuple(), tuple()),
        ([1], (1,)),
        ((1,), (1,)),
        ([1, 2], (1, 2)),
        ((1, 2), (1, 2)),
        ([[1, 2]], ((1, 2),)),
        (((1, 2),), ((1, 2),)),
        ([[1], [2]], ((1,), (2,))),
        (([1], [2]), ((1,), (2,))),
        (((1,), (2,)), ((1,), (2,))),
        ([[[1]]], (((1,),),)),
        ([[[1, 2], [3, 4]]], (((1, 2), (3, 4)),)),
    ],
)
def test_normalize_array(arr, expected):
    """Test normalize array."""
    assert normalize_array(arr) == expected
    arr = array(arr)
    assert normalize_array(arr) == expected



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""ZNE functionality for :method:`qiskit.primitives.BaseEstimator.__init__` method."""
from __future__ import annotations

from collections.abc import Callable, Iterable, Sequence
from functools import wraps

from numpy import ndarray
from qiskit.circuit import Parameter, QuantumCircuit
from qiskit.quantum_info.operators import SparsePauliOp

from ..zne_strategy import ZNEStrategy

Sequence.register(ndarray)


def zne_init(init: Callable) -> Callable:
    """Add ZNE functionality to :method:`qiskit.primitives.BaseEstimator.__init__`."""

    if not callable(init):
        raise TypeError("Invalid `init` argument, expected callable.")

    @wraps(init)
    def _zne_init(  # pylint: disable=too-many-arguments
        self,
        circuits: Iterable[QuantumCircuit] | QuantumCircuit | None = None,
        observables: Iterable[SparsePauliOp] | SparsePauliOp | None = None,
        parameters: Iterable[Iterable[Parameter]] | None = None,
        options: dict | None = None,
        zne_strategy: ZNEStrategy | None = None,
        **kwargs,
    ) -> None:
        if circuits is not None or observables is not None or parameters is not None:
            raise TypeError(
                "The BaseEstimator `circuits`, `observables`, `parameters` kwarg are "
                "deprecated as of Qiskit Terra 0.22.0. Use the 'run' method instead.",
            )
        self.zne_strategy = zne_strategy
        return init(self, options=options, **kwargs)

    return _zne_init



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Noise amplification strategy via unitary/inverse repetition."""

from __future__ import annotations

from warnings import warn

from numpy.random import Generator, default_rng
from qiskit import QuantumCircuit
from qiskit.dagcircuit import DAGCircuit

from ...utils.typing import isreal
from ..noise_amplifier import NoiseAmplifier


class FoldingAmplifier(NoiseAmplifier):
    """Interface for folding amplifier strategies."""

    def __init__(  # pylint: disable=super-init-not-called, duplicate-code, too-many-arguments
        self,
        sub_folding_option: str = "from_first",
        barriers: bool = True,
        random_seed: int | None = None,
        noise_factor_relative_tolerance: float = 1e-2,
        warn_user: bool = True,
    ) -> None:
        """
        Args:
            sub_folding_option: Specifies which gates are used for sub folding when
                ``noise_factor`` is not an odd integer. Can either be "from_last", "from_first", or
                "random".
            barriers: If True applies barriers when folding (e.g. to avoid simplification).
            random_seed: Random seed used for performing random sub-gate-folding.
            noise_factor_relative_tolerance: Relative allowed tolerance interval between
                ``noise_factor`` input and actual noise factor that was used for the amplification.
                If the discrepancy exceeds the tolerance, a warning is displayed.
            warn_user: Specifies whether user warnings are displayed.
        """
        self.warn_user: bool = warn_user
        self._set_sub_folding_option(sub_folding_option)
        self._set_barriers(barriers)
        self._prepare_rng(random_seed)
        self._set_noise_factor_relative_tolerance(noise_factor_relative_tolerance)

    ################################################################################
    ## PROPERTIES
    ################################################################################
    @property
    def options(self) -> dict:
        """Strategy options."""
        options: dict = self._init_options
        options.pop("warn_user", None)
        return options

    @property
    def warn_user(self) -> bool:
        """Option whether warnings are displayed."""
        return self._warn_user

    @warn_user.setter
    def warn_user(self, warn_user: bool) -> None:
        if not isinstance(warn_user, bool):
            raise TypeError(f"Expected boolean, received {type(warn_user)} instead.")
        self._warn_user: bool = warn_user

    @property
    def sub_folding_option(self) -> str:
        """Option for sub-gate-folding.

        Specifies which gates of the full circuit are used for sub folding when ``noise_factor`` is
        is not an odd integer. Can either be "from_last", "from_first", or "random".
        """
        return self._sub_folding_option

    def _set_sub_folding_option(self, sub_folding_option: str) -> None:
        if sub_folding_option not in ["from_last", "from_first", "random"]:
            raise ValueError(
                f"Sub-folding option must be a string of either 'from_last', 'from_first, or "
                f"'random'. Received {sub_folding_option} instead."
            )
        self._sub_folding_option: str = sub_folding_option

    @property
    def barriers(self) -> bool:
        """Option for whether to apply barriers when folding."""
        return self._barriers

    def _set_barriers(self, barriers: bool) -> None:
        self._barriers = bool(barriers)

    def _prepare_rng(self, seed: int | None = None) -> None:
        """Sets random number generator with seed."""
        if not isinstance(seed, (type(None), int)):
            raise TypeError("Random seed must be an integer or None.")
        self._rng: Generator = default_rng(seed)

    # TODO: add property getter
    def _set_noise_factor_relative_tolerance(self, tolerance: float) -> None:
        """Sets noise factor relative tolerance."""
        if not isreal(tolerance):
            raise TypeError("Noise factor relative tolerance must be real valued.")
        self._noise_factor_relative_tolerance: float = tolerance

    ################################################################################
    ## FOLDING METHODS
    ################################################################################
    @staticmethod
    def folding_to_noise_factor(folding: float) -> float:
        """Converts number of foldings to noise factor.

        Args:
            folding: The number of foldings.

        Returns:
            The corresponding noise factor.
        """
        return 2 * folding + 1

    def warn(self, *args, **kwargs):
        """Throws user warning if enabled."""
        if self.warn_user:
            warn(*args, **kwargs)

    def _validate_noise_factor(self, noise_factor):
        """Validates noise factor.

        Args:
            noise_factor: The noise amplification factor by which to amplify the circuit noise.

        Raises:
            ValueError: If the noise factor is smaller than 1.
        """
        if noise_factor < 1:
            raise ValueError(
                f"{self.name} expects a positive float noise_factor >= 1."
                f"Received {noise_factor} instead."
            )

    def _apply_barrier(self, circuit: QuantumCircuit, *registers) -> QuantumCircuit:
        """Apply barrier to specified registers if option is set."""
        if self._barriers:
            circuit.barrier(*registers)
        return circuit

    def _compute_folding_nums(self, noise_factor: float, num_instructions: int) -> tuple[int, int]:
        """Returns required number of full foldings and sub foldings.

        Args:
            noise_factor: The noise amplification factor by which to fold the circuit.
            num_instructions: The number of instructions to fold.

        Returns:
            A tuple containing the number of full-and sub-foldings.
        """
        if num_instructions == 0:
            self.warn("Noise amplification is not performed since none of the gates are folded.")
            return 0, 0
        num_foldings = round(num_instructions * (noise_factor - 1) / 2.0)
        closest_noise_factor: float = self.folding_to_noise_factor(num_foldings / num_instructions)
        relative_error = abs(closest_noise_factor - noise_factor) / noise_factor
        if relative_error > self._noise_factor_relative_tolerance:
            warn(
                "Rounding of noise factor: Foldings are performed with noise factor "
                f"{closest_noise_factor:.2f} instead of specified noise factor "
                f"{noise_factor:.2f} which amounts to a relative error of "
                f"{relative_error * 100:.2f}%."
            )
        num_full_foldings, num_sub_foldings = divmod(num_foldings, num_instructions)
        return num_full_foldings, num_sub_foldings

    ################################################################################
    ## IMPLEMENTATION
    ################################################################################
    # pylint: disable=useless-parent-delegation
    def amplify_dag_noise(
        self, dag: DAGCircuit, noise_factor: float
    ) -> DAGCircuit:  # pragma: no cover
        return super().amplify_dag_noise(dag, noise_factor)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from unittest.mock import Mock

from pytest import mark, raises

from zne.meta.call import zne_call


################################################################################
## DECORATOR
################################################################################
def test_zne_call():
    call = zne_call(Mock())
    with raises(TypeError):
        call("self", "circuits", "observables")


@mark.parametrize("call", [None, Ellipsis, True, 0, 1.0, 1j, "call"])
def test_type_error(call):
    with raises(TypeError):
        zne_call(call)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Docstrings utils module."""

from __future__ import annotations

SECTIONS = ["Args", "Returns", "Yields", "Raises"]  # TODO add header section


def insert_into_docstring(original_docstring: str, new_content_list: list[tuple[str, str]]) -> str:
    """Inserts new lines of content into existing docstring.

    Args:
        original_docstring: The docstring where to insert new lines.
        new_content_list: A list of tuples. The first element of a tuple specifies the section to
            append to, while the second element contains the new information to be added to that
            section, e.g.: [("Args", "param: example"), ("Raises", "Error: example")].

    Returns:
        Updated docstring.
    """
    docstring_lines = original_docstring.split("\n")
    docstring_lines = remove_all_leading_spaces_in_docstring(docstring_lines)
    for section, new_content in new_content_list:
        validate_section(section)
        docstring_lines = insert_into_docstring_lines(docstring_lines, new_content, section)
    return "\n".join(docstring_lines)


def remove_all_leading_spaces_in_docstring(docstring_lines: list[str]) -> list[str]:
    """Removes all common leading spaces in each line.

    Args:
        docstring_lines: A list of docstring lines.

    Returns:
        Docstring lines with leading spaces removed.
    """
    num_leading_spaces = get_num_leading_spaces(docstring_lines)
    docstring_lines[1:] = [line[num_leading_spaces:] for line in docstring_lines[1:]]
    return docstring_lines


def get_num_leading_spaces(docstring_lines: list[str]) -> int:
    """Returns the number of leading spaces in docstring.

    Args:
        docstring_lines: A list of docstring lines.

    Returns:
        The number of leading spaces.

    Raises:
        RuntimeError: If docstring is empty or one-liner.
    """
    for line in docstring_lines[1:]:
        if len(line) > 0:
            return len(line) - len(line.lstrip(" "))
    raise ValueError("Functionality not supported for single-line docstrings.")


def validate_section(section: str) -> None:
    """Validates provided docstring section.

    Args:
        section: The docstring section to be validated.

    Raises:
        ValueError: If section not a valid docstring section.
    """
    if section not in SECTIONS:
        raise ValueError(f"section has to be one of {SECTIONS}. Received {section} instead.")


def insert_into_docstring_lines(
    docstring_lines: list[str], new_content: str, section: str
) -> list[str]:
    """Inserts new entry into list of docstring lines.

    Args:
        docstring_lines: A list of docstring lines.
        new_content: The new content.
        section: The section to append to.

    Returns:
        The updated list of docstring lines.
    """
    section_exists = check_section_exists_in_docstring(docstring_lines, section)
    insert_location = get_insert_location_in_docstring(docstring_lines, section)
    if not section_exists:
        docstring_lines, insert_location = insert_new_section_into_docstring(
            docstring_lines, insert_location, section
        )
    docstring_lines, insert_location = insert_new_content_into_docstring(
        docstring_lines, insert_location, new_content
    )
    return docstring_lines


def check_section_exists_in_docstring(docstring_lines: list[str], section: str) -> bool:
    """Checks whether section already exists in docstring.

    Args:
        docstring_lines: A list of docstring lines.
        section: The section to append to.

    Returns:
        True if section already exists and False otherwise.
    """
    for line in docstring_lines:
        if section in line:
            return True
    return False


def get_insert_location_in_docstring(docstring_lines: list[str], section: str) -> int:
    """Returns index at which to insert into docstring.

    Args:
        docstring_lines: A list of docstring lines.
        section: The section to append to.

    Returns:
        The index.
    """
    section_idx = SECTIONS.index(section)
    truncated_sections = SECTIONS[section_idx + 1 :]
    for idx, line in enumerate(docstring_lines):
        if any(section in line for section in truncated_sections):
            return idx - 1
    return len(docstring_lines) - 1


def insert_new_section_into_docstring(
    docstring_lines: list[str], insert_location: int, section: str
) -> tuple[list[str], int]:
    """Inserts a new section heading into docstring.

    Args:
        docstring_lines: A list of docstring lines.
        insert_location: The index at which to insert.
        section: The section to append to.

    Returns:
        A tuple containing the updated list of docstring lines and insert location.
    """
    docstring_lines.insert(insert_location, "")
    docstring_lines.insert(insert_location + 1, section + ":")
    return docstring_lines, insert_location + 2


def insert_new_content_into_docstring(
    docstring_lines: list[str], insert_location: int, new_content: str
) -> tuple[list[str], int]:
    """Inserts a new line of content into docstring.

    Args:
        docstring_lines: A list of docstring lines.
        insert_location: The index at which to insert.
        new_content: The new content.

    Returns:
        A tuple containing the updated list of docstring lines and insert location.
    """
    new_content_lines = new_content.split("\n")
    counter = 0
    for counter, line in enumerate(new_content_lines):
        # TODO infer tab size instead of hardcoding it
        num_leading_spaces = 4 if counter == 0 else 8
        docstring_lines.insert(insert_location + counter, " " * num_leading_spaces + line)
    return docstring_lines, insert_location + counter + 1



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.


from itertools import product
from unittest.mock import Mock, patch

from pytest import mark
from qiskit.circuit.random import random_circuit
from qiskit.quantum_info.operators import Operator
from qiskit.transpiler import TransformationPass

from zne import NOISE_AMPLIFIER_LIBRARY


@mark.parametrize(
    "NoiseAmplifier",
    NOISE_AMPLIFIER_LIBRARY.values(),
    ids=NOISE_AMPLIFIER_LIBRARY.keys(),
)
class TestNoiseAmplifier:
    @mark.parametrize(
        "circuit, noise_factor",
        cases := tuple(
            product(
                [
                    random_circuit(5, 3, seed=0),
                    random_circuit(5, 3, seed=5),
                    random_circuit(5, 6, seed=66),
                    random_circuit(5, 6, seed=1081),
                ],
                [1, 3, 5],
            )
        ),
        ids=[f"{type(c).__name__}<{len(c)}>-{nf}" for c, nf in cases],
    )
    @mark.filterwarnings("ignore::UserWarning")
    def test_circuit_equivalence(self, NoiseAmplifier, circuit, noise_factor):
        noisy_circuit = NoiseAmplifier().amplify_circuit_noise(circuit.copy(), noise_factor)
        assert Operator(noisy_circuit).equiv(Operator(circuit))

    ################################################################################
    ## TESTS
    ################################################################################
    @mark.parametrize("noise_factor", [1, 1.2, 2.4, -3.14])
    def test_build_transpiler_pass(self, NoiseAmplifier, noise_factor):
        DAG = f"DAG:{noise_factor}"
        AMP_DAG = f"AMP_DAG:{noise_factor}"
        noise_amplifier = NoiseAmplifier()
        noise_amplifier.amplify_dag_noise = Mock(return_value=AMP_DAG)
        transpiler_pass = noise_amplifier.build_transpiler_pass(noise_factor)
        assert isinstance(transpiler_pass, TransformationPass)
        assert AMP_DAG == transpiler_pass.run(DAG)
        noise_amplifier.amplify_dag_noise.assert_called_once_with(DAG, noise_factor)

    @mark.parametrize("noise_factor", [1, 1.2, 2.4, -3.14])
    def test_build_pass_manager(self, NoiseAmplifier, noise_factor):
        target = "zne.noise_amplification.noise_amplifier"
        target_pass_builder = target + ".NoiseAmplifier.build_transpiler_pass"
        target_pass_manager_class = target + ".PassManager"
        TRANSPILER_PASS = f"<NoiseAmplificationPass:{noise_factor}>"
        with patch(target_pass_builder, return_value=TRANSPILER_PASS) as pass_builder, patch(
            target_pass_manager_class
        ) as pass_manager_class:
            _ = NoiseAmplifier().build_pass_manager(noise_factor)
        pass_builder.assert_called_once_with(noise_factor)
        pass_manager_class.assert_called_once_with(TRANSPILER_PASS)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Interface for noise amplification strategies."""

from abc import ABC, abstractmethod

from qiskit.circuit import QuantumCircuit
from qiskit.converters import circuit_to_dag, dag_to_circuit
from qiskit.dagcircuit import DAGCircuit
from qiskit.transpiler import PassManager, TransformationPass

from ..immutable_strategy import ImmutableStrategy


class NoiseAmplifier(ImmutableStrategy, ABC):
    """Interface for noise amplification strategies."""

    @abstractmethod
    def amplify_circuit_noise(
        self, circuit: QuantumCircuit, noise_factor: float
    ) -> QuantumCircuit:  # pragma: no cover
        """Noise amplification strategy over :class:`~qiskit.circuit.QuantumCircuit`.

        Args:
            circuit: The original quantum circuit.
            noise_factor: The noise amplification factor by which to amplify the circuit noise.

        Returns:
            The noise amplified quantum circuit
        """
        dag: DAGCircuit = circuit_to_dag(circuit)
        dag = self.amplify_dag_noise(dag, noise_factor)
        return dag_to_circuit(dag)

    @abstractmethod
    def amplify_dag_noise(
        self, dag: DAGCircuit, noise_factor: float
    ) -> DAGCircuit:  # pragma: no cover
        """Noise amplification strategy over :class:`~qiskit.dagcircuit.DAGCircuit`.

        Args:
            dag: The original dag circuit.
            noise_factor: The noise amplification factor by which to amplify the circuit noise.

        Returns:
            The noise amplified dag circuit
        """
        circuit: QuantumCircuit = dag_to_circuit(dag)
        circuit = self.amplify_circuit_noise(circuit, noise_factor)
        return circuit_to_dag(circuit)

    def build_transpiler_pass(self, noise_factor: float) -> TransformationPass:
        """Builds transpiler pass to perform noise amplification as specified in the strategy.

        Args:
            noise_factor: The noise factor used to amplify circuit noise.

        Returns:
            NoiseAmplificationPass: Instance of TransformationPass.
        """

        class NoiseAmplificationPass(TransformationPass):
            """Noise amplification transpiler pass."""

            def __init__(self, noise_amplifier: NoiseAmplifier) -> None:
                super().__init__()
                self._noise_amplifier: NoiseAmplifier = noise_amplifier

            @property
            def noise_amplifier(self) -> NoiseAmplifier:
                """Underlying noise amplifier."""
                return self._noise_amplifier

            def run(self, dag: DAGCircuit) -> DAGCircuit:
                """Run a pass on the DAGCircuit."""
                return self.noise_amplifier.amplify_dag_noise(dag, noise_factor)

        return NoiseAmplificationPass(noise_amplifier=self)

    def build_pass_manager(self, noise_factor: float) -> PassManager:
        """Builds a pass manager holding a single transpiler pass for noise amplification.

        Args:
            noise_factor: The noise factor used to amplify circuit noise.

        Returns:
            PassManager: Wrapper for :class:`NoiseAmplificationPass`.
        """
        transpiler_pass: TransformationPass = self.build_transpiler_pass(noise_factor)
        return PassManager(transpiler_pass)


# TODO: deprecate
class CircuitNoiseAmplifier(NoiseAmplifier):  # pragma: no cover
    """Interface for noise amplification strategies over :class:`~qiskit.dagcircuit.DAGCircuit`."""

    # pylint: disable=useless-parent-delegation
    def amplify_dag_noise(self, dag: DAGCircuit, noise_factor: float) -> DAGCircuit:
        return super().amplify_dag_noise(dag, noise_factor)


# TODO: deprecate
class DAGNoiseAmplifier(NoiseAmplifier):  # pragma: no cover
    """Interface for noise amplification strategies over :class:`~qiskit.dagcircuit.DAGCircuit`."""

    # pylint: disable=useless-parent-delegation
    def amplify_circuit_noise(self, circuit: QuantumCircuit, noise_factor: float) -> QuantumCircuit:
        return super().amplify_circuit_noise(circuit, noise_factor)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Estimator Job with embeded ZNE parsing."""

from __future__ import annotations

from typing import Any

from qiskit.primitives import EstimatorResult
from qiskit.providers import JobV1 as Job

from ..zne_strategy import ZNEStrategy


class ZNEJob(Job):
    """Estimator Job with embeded ZNE parsing."""

    def __init__(  # pylint: disable=super-init-not-called
        self,
        base_job: Job,
        zne_strategy: ZNEStrategy,
        target_num_experiments: int,
    ) -> None:
        self._base_job: Job = base_job
        self._zne_strategy: ZNEStrategy = zne_strategy
        self._target_num_experiments: int = target_num_experiments

    ################################################################################
    ## PROPERTIES
    ################################################################################
    @property
    def base_job(self) -> Job:
        """Base Job for noisy results."""
        return self._base_job

    @property
    def zne_strategy(self) -> ZNEStrategy:
        """ZNE strategy object."""
        return self._zne_strategy

    @property
    def target_num_experiments(self) -> int:
        """Expected number of results after ZNE is performed."""
        return self._target_num_experiments

    ################################################################################
    ## ZNE
    ################################################################################
    def result(self) -> EstimatorResult:
        """Return the results of the job."""
        result: EstimatorResult = self.base_job.result()
        if self.zne_strategy.performs_zne:
            result = self.zne_strategy.mitigate_noisy_result(result)
        if result.num_experiments != self.target_num_experiments:
            # TODO: consider warning instead -> should be in integration tests
            raise RuntimeError(
                "Number of experiments in EstimatorResult object does not match "
                "the requested number of experiments."
            )
        return result

    ################################################################################
    ## NESTED JOB DELEGATION
    ################################################################################
    _ZNE_ATTRIBUTES: set = {
        "base_job",
        "_base_job",
        "zne_strategy",
        "_zne_strategy",
        "target_num_experiments",
        "_target_num_experiments",
        "result",
    }

    def __getattribute__(self, name: str) -> Any:
        if name == "_ZNE_ATTRIBUTES" or name in self._ZNE_ATTRIBUTES:
            return super().__getattribute__(name)
        return getattr(self.base_job, name)

    def __setattr__(self, name: str, value: Any) -> None:
        if name in self._ZNE_ATTRIBUTES:
            return super().__setattr__(name, value)
        return setattr(self.base_job, name, value)

    def __delattr__(self, name: str) -> None:
        if name in self._ZNE_ATTRIBUTES:
            return super().__delattr__(name)
        return delattr(self.base_job, name)

    ################################################################################
    ## DUMMY ABSTRACT METHOD IMPLEMENTATION
    ################################################################################
    def submit(self):
        """Submit the job to the backend for execution."""
        return self.base_job.submit()

    def status(self):
        """Return the status of the job, among the values of ``JobStatus``."""
        return self.base_job.status()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from copy import copy, deepcopy

from zne.utils.unset import UNSET, UnsetType


class TestUnsetType:
    """Test `UnsetType` class."""

    def test_singleton(self):
        """Test that class is a singleton."""
        assert UnsetType() is UnsetType()
        assert UNSET is UnsetType()

    def test_bool(self):
        """Test that instances evaluate to `False`."""
        assert not UNSET
        assert not UnsetType()

    def test_eq(self):
        """Test equality."""
        assert UnsetType() == UnsetType()
        assert UNSET == UnsetType()

    def test_repr(sefl):
        """Test string representation."""
        assert str(UNSET) == "UNSET"
        assert repr(UNSET) == "UNSET"

    def test_copy(self):
        """Test copy."""
        assert copy(UNSET) is UNSET

    def test_deepcopy(self):
        """Test deepcopy."""
        assert deepcopy(UNSET, memo := {}) is UNSET
        assert not memo



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Class constant descriptor."""

from __future__ import annotations

from copy import deepcopy
from typing import Any


class classconstant:  # pylint: disable=invalid-name

    """Class constant descriptor.

    Args:
        value: the constant value to store at the class level.
        copy: whether to return a deepcopy of the value or the original reference.

    Note: returning constant values by reference is more performant but opens the up
    the possibility to mutating them. For this reason, if the value type is not
    immutable, setting `copy` to `True` is recommended.
    """

    def __init__(self, value: Any, copy: bool = False) -> None:
        self.name: str = ""
        self.value: Any = value
        self.copy: bool = copy

    def __set_name__(self, owner: type, name: str) -> None:
        self.name = name

    def __get__(self, obj: object, objtype: type = None) -> Any:
        return deepcopy(self.value) if self.copy else self.value

    def __set__(self, obj: object, value: Any) -> None:
        raise AttributeError(f"Class constant '{self.name}' cannot be assigned.")

    def __delete__(self, obj: object) -> None:
        raise AttributeError(f"Class constant '{self.name}' cannot be deleted.")



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from copy import copy, deepcopy
from typing import Any
from unittest.mock import Mock

from pytest import mark, raises

from zne.utils.unset import UNSET
from zne.utils.validation import quality


################################################################################
## AUXILIARY
################################################################################
class EqualCopies:
    """Auxiliary wrapper class to check that deepcopies are equal but not the same object."""

    def __init__(self, value: Any = None) -> None:
        self.value = value

    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, self.__class__):
            return False
        return self.value == __o.value


################################################################################
## TESTS
################################################################################
class TestQualityInit:
    """Test `quality` descriptor init."""

    def test_defaults(self):
        q = quality()
        assert q.fval is None
        assert q.feff is None
        assert q.default is UNSET
        assert q.null is UNSET

    def test_custom(self):
        fval = Mock()
        feff = Mock()
        default = EqualCopies("default")
        null = EqualCopies("null")
        q = quality(fval, feff, default=default, null=null)
        assert q.fval is fval
        assert q.feff is feff
        assert q.default == default
        assert q.null == null

    def test_keyword_only(self):
        with raises(TypeError):
            quality(None, None, "default")

    def test_uset_name(self):
        q = quality()
        assert q.name is None
        assert q.private_name is None


class TestQualityCall:
    """Test `quality` descriptor call."""

    @mark.parametrize(
        "keys",
        [
            ("fval",),
            ("feff",),
            ("fval", "feff"),
        ],
    )
    def test_mutable(self, keys):
        q = quality()
        kwargs = {k: Mock() for k in keys}
        qq = q(**kwargs)
        assert q is not qq
        for key, value in kwargs.items():
            assert getattr(qq, key) is value
            assert getattr(q, key) is not value

    @mark.parametrize(
        "keys",
        [
            ("default",),
            ("null",),
            ("default", "null"),
        ],
    )
    def test_immutable(self, keys):
        q = quality()
        kwargs = {k: EqualCopies(k) for k in keys}
        qq = q(**kwargs)
        assert q is not qq
        for key, value in kwargs.items():
            assert getattr(qq, key) == value
            assert getattr(q, key) != value

    @mark.parametrize("name", [None, "attr", "name", "foo", "bar"])
    def test_name(self, name):
        q = quality()
        q._name = name
        qq = q()
        assert q is not qq
        assert qq.name == q.name
        assert qq.private_name == q.private_name


@mark.parametrize("name", [None, "attr", "name", "foo", "bar"])
class TestQualityCopy:
    """Test `quality` descriptor copy logic."""

    def test_copy(self, name):
        fval = Mock()
        feff = Mock()
        default = EqualCopies("default")
        null = EqualCopies("null")
        descriptor = quality(fval, feff, default=default, null=null)
        descriptor._name = name
        duplicate = copy(descriptor)
        assert descriptor is not duplicate
        assert descriptor.fval is duplicate.fval
        assert descriptor.feff is duplicate.feff
        assert descriptor.default == duplicate.default
        assert descriptor.null == duplicate.null
        assert descriptor.name == duplicate.name
        assert descriptor.private_name == duplicate.private_name

    def test_deepcopy(self, name):
        fval = EqualCopies("fval")
        feff = EqualCopies("feff")
        default = EqualCopies("default")
        null = EqualCopies("null")
        descriptor = quality(fval, feff, default=default, null=null)
        descriptor._name = name
        duplicate = deepcopy(descriptor, memo := {})

        assert descriptor is not duplicate
        assert descriptor.fval is not duplicate.fval
        assert descriptor.fval == duplicate.fval
        assert descriptor.feff is not duplicate.feff
        assert descriptor.feff == duplicate.feff
        assert descriptor.default is not duplicate.default
        assert descriptor.default == duplicate.default
        assert descriptor.null is not duplicate.null
        assert descriptor.null == duplicate.null
        assert descriptor.name == duplicate.name
        assert descriptor.private_name == duplicate.private_name

        memo_values = [fval, feff, default, null]
        memo_values += [v.__dict__ for v in memo_values]
        for value in memo.values():
            if isinstance(value, quality):
                assert value is duplicate
            elif isinstance(value, list):
                assert descriptor in value
                for v in memo_values:
                    assert v in value
            else:
                assert value in memo_values


class TestQualityProperties:
    """Test `quality` descriptor properties."""

    @mark.parametrize("property", ["default", "null", "name", "private_name"])
    def test_immutable(self, property):
        descriptor = quality()
        with raises(AttributeError):
            setattr(descriptor, property, ...)

    @mark.parametrize("property", ["default", "null"])
    def test_copy(self, property):
        p = EqualCopies(property)
        descriptor = quality(**{property: p})
        assert getattr(descriptor, property) is not p
        assert getattr(descriptor, property) == p


class TestQualityDescriptor:
    """Test `quality` descriptor protocol."""

    @mark.parametrize("attr", [None, "attr", "name", "foo", "bar"])
    def test_set_name(self, attr):
        """Test `quality` descriptor `__set_name__` logic."""
        descriptor = quality()
        type("Dummy", (), {attr: descriptor})
        assert descriptor.name == attr
        assert descriptor.private_name == attr if attr is None else "__quality_" + attr

    def test_get(self):
        """Test `quality` descriptor `__get__` logic."""
        descriptor = quality()
        Dummy = type("Dummy", (), {"q": descriptor})
        assert Dummy.q is descriptor

    @mark.parametrize(
        "kwargs",
        [
            {"default": UNSET, "null": UNSET},
            {"default": EqualCopies("default"), "null": UNSET},
            {"default": UNSET, "null": EqualCopies("null")},
            {"default": EqualCopies("default"), "null": EqualCopies("null")},
        ],
    )
    class TestSet:
        """Test `quality` descriptor `__set__` logic."""

        def test_unset(self, kwargs):
            default = kwargs.get("default")
            null = kwargs.get("null")
            descriptor = quality(**kwargs)
            Dummy = type("Dummy", (), {"q": descriptor})
            d = Dummy()
            expected = None if default is UNSET and null is UNSET else default or null
            assert d.q is not UNSET
            assert d.q == expected
            assert d.q == getattr(d, descriptor.private_name)

        def test_none(self, kwargs):
            default = kwargs.get("default")
            null = kwargs.get("null")
            descriptor = quality(**kwargs)
            Dummy = type("Dummy", (), {"q": descriptor})
            d = Dummy()
            d.q = None
            expected = None if default is UNSET and null is UNSET else null or default
            assert d.q is not UNSET
            assert d.q == expected
            assert d.q == getattr(d, descriptor.private_name)

        def test_ellipsis(self, kwargs):
            default = kwargs.get("default")
            null = kwargs.get("null")
            descriptor = quality(**kwargs)
            Dummy = type("Dummy", (), {"q": descriptor})
            d = Dummy()
            d.q = ...
            expected = ... if default is UNSET and null is UNSET else default or null
            assert d.q is not UNSET
            assert d.q == expected
            assert d.q == getattr(d, descriptor.private_name)

        def test_set(self, kwargs):
            descriptor = quality(**kwargs)
            Dummy = type("Dummy", (), {"q": descriptor})
            d = Dummy()
            q = Mock()
            d.q = q
            assert d.q is q
            assert d.q is getattr(d, descriptor.private_name)

    def test_delete(self):
        """Test `quality` descriptor `__delete__` logic."""
        descriptor = quality()
        Dummy = type("Dummy", (), {"q": descriptor})
        d = Dummy()
        assert not hasattr(d, descriptor.private_name)
        del d.q  # Does not raise error
        q = Mock()
        d.q = q
        assert getattr(d, descriptor.private_name) is q
        del d.q
        assert not hasattr(d, descriptor.private_name)

    def test_fval(self):
        """Test `quality` descriptor `fval` attribute."""
        fval = Mock()
        descriptor = quality(fval=fval)
        Dummy = type("Dummy", (), {"q": descriptor})
        d = Dummy()
        q = Mock()
        d.q = q
        fval.assert_called_once_with(d, q)

    def test_feff(self):
        """Test `quality` descriptor `feff` attribute."""
        feff = Mock()
        descriptor = quality(feff=feff)
        Dummy = type("Dummy", (), {"q": descriptor})
        d = Dummy()
        q = Mock()
        d.q = q
        feff.assert_called_once_with(d)
        feff.reset_mock()
        del d.q
        feff.assert_called_once_with(d)


class TestQualityDecorators:
    """Test `quality` descriptor decorators."""

    def test_validator(self):
        """Test validator decorator."""
        q = quality()
        fval = Mock()
        qq = q.validator(fval)
        assert qq is not q
        assert qq.fval is fval
        assert qq.fval is not q.fval
        assert qq.feff is q.feff
        assert qq.default == q.default
        assert qq.null == q.null
        assert qq.name == q.name
        assert qq.private_name == q.private_name

    def test_side_effect(self):
        """Test side effect decorator."""
        q = quality()
        feff = Mock()
        qq = q.side_effect(feff)
        assert qq is not q
        assert qq.fval is q.fval
        assert qq.feff is feff
        assert qq.feff is not q.feff
        assert qq.default == q.default
        assert qq.null == q.null
        assert qq.name == q.name
        assert qq.private_name == q.private_name



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""ZNE functionality for :method:`qiskit.primitives.BaseEstimator._run` method."""

from __future__ import annotations

from collections.abc import Callable
from functools import wraps

from qiskit.circuit import QuantumCircuit
from qiskit.providers import JobV1 as Job
from qiskit.quantum_info.operators.base_operator import BaseOperator

from ..zne_strategy import ZNEStrategy
from .job import ZNEJob


def zne_run(run: Callable) -> Callable:
    """Add ZNE functionality to :method:`qiskit.primitives.BaseEstimator._run`."""

    if not callable(run):
        raise TypeError("Invalid `run` argument, expected callable.")

    @wraps(run)
    def _zne_run(
        self,
        circuits: tuple[QuantumCircuit, ...],
        observables: tuple[BaseOperator, ...],
        parameter_values: tuple[tuple[float, ...], ...],
        zne_strategy: ZNEStrategy | None = ...,  # type: ignore
        **run_options,
    ) -> Job:
        # Strategy
        if zne_strategy is Ellipsis:
            zne_strategy = self.zne_strategy
        elif zne_strategy is None:
            zne_strategy = ZNEStrategy.noop()
        elif not isinstance(zne_strategy, ZNEStrategy):
            raise TypeError("Invalid zne_strategy object, expected ZNEStrategy.")

        # ZNE
        target_num_experiments: int = len(circuits)
        if zne_strategy.performs_noise_amplification:
            circuits = zne_strategy.build_noisy_circuits(circuits)
            observables = zne_strategy.map_to_noisy_circuits(observables)
            parameter_values = zne_strategy.map_to_noisy_circuits(parameter_values)

        # Job
        job: Job = run(
            self,
            circuits=circuits,
            observables=observables,
            parameter_values=parameter_values,
            **run_options,
        )
        return ZNEJob(job, zne_strategy, target_num_experiments)

    return _zne_run



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Zero Noise Extrapolation (ZNE) prototype for error mitigation on
Qiskit's Estimator primitive."""

from .extrapolation import EXTRAPOLATOR_LIBRARY
from .meta import ZNEJob, zne
from .noise_amplification import NOISE_AMPLIFIER_LIBRARY
from .zne_strategy import ZNEStrategy

__copyright__ = "(C) Copyright IBM 2022"
__version__ = "1.3.0"


__all__ = [
    "__copyright__",
    "__version__",
    "zne",
    "ZNEJob",
    "ZNEStrategy",
    "EXTRAPOLATOR_LIBRARY",
    "NOISE_AMPLIFIER_LIBRARY",
]



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Extrapolation library."""

from .exponential_extrapolator import (
    BiExponentialExtrapolator,
    ExponentialExtrapolator,
    MonoExponentialExtrapolator,
    MultiExponentialExtrapolator,
)
from .extrapolator import Extrapolator, OLSExtrapolator, ReckoningResult
from .polynomial_extrapolator import (
    CubicExtrapolator,
    LinearExtrapolator,
    PolynomialExtrapolator,
    QuadraticExtrapolator,
    QuarticExtrapolator,
)

EXTRAPOLATOR_LIBRARY = {
    cls.__name__: cls
    for cls in (
        PolynomialExtrapolator,
        LinearExtrapolator,
        QuadraticExtrapolator,
        CubicExtrapolator,
        QuarticExtrapolator,
        MultiExponentialExtrapolator,
        ExponentialExtrapolator,
        MonoExponentialExtrapolator,
        BiExponentialExtrapolator,
    )
}

__all__ = [
    "Extrapolator",
    "OLSExtrapolator",
    "ReckoningResult",
    *EXTRAPOLATOR_LIBRARY.keys(),
]



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""ZNE meta programming capabilities.

Enables injecting error mitigation functionality to classes implementing the
:class:`qiskit.primitives.BaseEstimator` interface.
"""

from __future__ import annotations

from qiskit.primitives import BaseEstimator

from zne.meta.call import zne_call
from zne.meta.init import zne_init
from zne.meta.run import zne_run
from zne.zne_strategy import ZNEStrategy


def zne(cls: type) -> type:  # TODO: integration tests
    """Add ZNE functionality to input class.

    Args:
        cls: class implementing the :class:`qiskit.primitives.BaseEstimator` interface.

    Returns:
        A subclass of the input class extended with ZNE functionality.
    """
    if not isinstance(cls, type) or not issubclass(cls, BaseEstimator):
        raise TypeError("Invalid class, does not implement the BaseEstimator interface.")
    namespace = {
        "__init__": zne_init(cls.__init__),  # type: ignore  # TODO: update docstring
        "__call__": zne_call(cls.__call__),  # TODO: deprecate
        "_run": zne_run(cls._run),  # type: ignore  # pylint: disable=protected-access
        "zne_strategy": property(_get_zne_strategy, _set_zne_strategy),
    }
    return type(f"ZNE{cls.__name__}", (cls,), namespace)


def _get_zne_strategy(self) -> ZNEStrategy:
    """ZNE strategy for error mitigation."""
    try:
        return self._zne_strategy  # pylint: disable=protected-access
    except AttributeError:
        self.zne_strategy = None
        return self.zne_strategy


def _set_zne_strategy(self, zne_strategy: ZNEStrategy | None) -> None:
    if zne_strategy is None:
        zne_strategy = ZNEStrategy.noop()
    elif not isinstance(zne_strategy, ZNEStrategy):
        raise TypeError("Invalid zne_strategy object, expected ZNEStrategy.")
    self._zne_strategy = zne_strategy  # pylint: disable=protected-access



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Strategy design pattern tools.

This module provides support for strategy classes, which are similar in format and
spirit to those produced by the builtin :class:`dataclass` decorator. In this case,
the relevant information is extracted from the class' ``__init__`` method signature
rather than from class fields. Said class signature specifies the so called 'settings'
for the strategy, where arguments whose name begins with a leading ``_`` are excluded
from the settings namespace. Although not currently enforced, it is strongly suggested
for these excluded arguments to be optional (i.e. developers should provide defaults).

Init's arguments are referred to as 'init args', and are not subject to any namespace
exclusions. Similarly, construction-time init args (including defaults) are referred to
as 'original args': they are merely a reflection of the values of all arguments passed
to ``__init__`` when instantiating the given strategy object.

Currently, init arguments of kind POSITIONAL_ONLY, VAR_POSITIONAL, and VAR_KEYWORD
are disallowed. See https://docs.python.org/3/library/inspect.html#inspect.Parameter.kind
for more information.

Strategy classes include the following specific mehods and fields:
    01. ``__init_subclass__``: to handle subclassing.
    02. ``__new__``: to save the original args at construction-time.
    03. ``__getattr__``: to retrieve original args via ``getattr`` (i.e. virtual attrs).
    04. ``__repr__``: a string representation of the strategy instance.
    05. ``__eq__``: to compare strategy instances.
    06. ``NAME``: the name of the strategy (i.e. class name).
    07. ``INIT_NAMESPACE``: a tuple of the strategy's settings names.
    08. ``SETTINGS_NAMESPACE``: a tuple of the strategy's settings names.
    09. ``settings``: a dictionary of settings and their corresponding values/states.
    10. ``init_args``: a dictionary of init arguments and their corresponding values/states.
    11. ``original_args``: a copy of the construction-time init args.
    12. ``is_facade``: to check is current class is a facade of a given ancestor.
    13. ``replicate``: to build replicas of any strategy instance with updated settings.

Out of these, {01, 02, 03} can extend and be extended, and {04, 05} can be overriden;
all others should be left unmodified.

Classes inheriting from a :class:`strategy` class, are strategy classes
themselves as well; with settings either inherited or provided by a new
``__init__`` method. This means that they do not need to be re-decorated.

Generally speaking, strategies are mutable, meaning that their settings
can be updated at runtime. The implicit assumption made by this decorator is
that the current state of all settings listed in ``SETTINGS_NAMESPACE`` can be
accessed through a ``getattr`` call. If any of these settings is not accessible
that way the corresponding (construction-time) init arg will be retrieved instead,
possibly leading to conflicts and runtime errors if mutated after instantiation.
If certain settings cannot be accessed by design, the suggested solution is
removing said settings from the settings namespace as explained before, or adding a
dummy property that retrieves their state (i.e. getter).

Oftentimes, developers will need their strategies to be immutable after instantiation.
To this end, the :class:`~zne.utils.strategy.strategy` decorator class accepts an
optional keyword argument ``frozen`` which, if set to ``True`` (default ``False``),
will forbid all ``setattr`` calls, and automatically return deepcopies of mutable
settings. In this scenario, settings will still be declared through the class'
``__init__`` method signature, while the body could be left blank (i.e. ``pass``),
hold the appropriate docstring, or used for validation among other things.

Replicas of any strategy instance can be produced by calling their
:method:`~zne.utils.strategy.strategy.replicate` method with updated settings
as keyword arguments. This is especially useful when dealing with frozen
(i.e. immutable) strategies.
"""

from __future__ import annotations

from abc import ABC
from copy import deepcopy
from inspect import Parameter, signature
from typing import Any

from .classconstant import classconstant
from .unset import UNSET


################################################################################
## CLASS DECORATOR
################################################################################
class strategy:  # pylint: disable=invalid-name
    """Strategy class decorator.

    Applies a strategy base class and its corresponding methods to the decorated class.

    This is done by dynamically subclassing from both the corresponding strategy base class
    (i.e. mutable or frozen), and the pre-decorated class as declared by the developer.
    The base strategy class will have precedence in the resulting class' mro; therefore,
    the resulting strategy class will extend the ``__init_subclass__``, ``__new__``, and
    ``__getattr__`` methods from the pre-decorated class. Additionally, in order to allow
    redefinition of the base strategy class' ``__init__`` (dummy), ``__repr__`` and ``__eq__``
    methods, if these are found in the pre-decorated class' dictionary, they will be bridged
    to the resulting class; effectively overriding the defaults. Similarly, ``__module__``,
    ``__name__``, ``__annotations__``, and ``__doc__``, will also be bridged from the
    pre-decorated class as is standard in decorators.

    Args:
        frozen: Whether to inherit from ``_BaseStrategy`` or ``_FrozenStrategy``.

    Example:
        >>> @strategy
        >>> class cls:
        >>>     def __init__(self, a, b, _c=None):
        >>>         pass
        >>> cls.mro()
        [__main__.cls, zne.utils.strategy._BaseStrategy, __main__.cls, object]

        >>> @strategy(frozen=True)
        >>> class cls:
        >>>     def __init__(self, a, b, _c=None):
        >>>         pass
        >>> cls.mro()
        [__main__.cls,
        zne.utils.strategy._FrozenStrategy,
        zne.utils.strategy._BaseStrategy,
        __main__.cls,
        object]
    """

    __slots__ = ("_frozen",)  # TODO: add `abstract` option

    OVERRIDING_NAMESPACE = (
        "__module__",
        "__annotations__",
        "__doc__",
        "__init__",
        "__repr__",
        "__eq__",
    )

    def __new__(cls, /, target: type = None, **kwargs) -> type | strategy:  # type: ignore
        """Return a decorated class or a strategy object capable of decorating."""
        self = super().__new__(cls)
        if target is None:
            return self  # Note: init auto-triggered for `strategy` return types
        return self(target, **kwargs)

    def __init__(self, /, *_, frozen: bool = False, **__) -> None:
        """Save decorator configuration."""
        self.frozen: bool = frozen

    def __call__(self, /, target: type, *, frozen: bool | None = None, **__) -> type:
        """Decorate target class."""
        frozen = self.frozen if frozen is None else frozen
        BaseStrategy = _FrozenStrategy if frozen else _BaseStrategy
        overriding_attrs = {
            attr: value
            for attr in self.OVERRIDING_NAMESPACE
            if (value := target.__dict__.get(attr, UNSET)) is not UNSET
        }
        return type(target.__name__, (BaseStrategy, target), overriding_attrs)

    @property
    def frozen(self) -> bool:
        """Frozen strategy."""
        return getattr(self, "_frozen", False)

    @frozen.setter
    def frozen(self, value: bool) -> None:
        self._frozen: bool = bool(value)


################################################################################
## AUXILIARY
################################################################################
def _pack_init_args(obj: object, *args, **kwargs) -> dict[str, Any]:
    """Packs object init options as a dict, binding missing kwargs with defaults.

    Args:
        obj: the object whose `__init__` method to inspect.
        args, kwargs: the particular values passed on to the `__init__` method.

    Returns:
        A dictionary sorted by signature, where keys are argument names,
        and values are either those provided or the `__init__` defaults.

    Note:
        Dictionaries keep insertion order as a langauage feature from Python 3.7.
        https://stackoverflow.com/questions/39980323/are-dictionaries-ordered-in-python-3-6#answers
    """
    init_signature = signature(obj.__init__)  # type: ignore
    bound_signature = init_signature.bind(*args, **kwargs)
    bound_signature.apply_defaults()
    bound_args = bound_signature.arguments  # Note: type `OrderedDict`
    return dict(bound_args)


def _infer_init_namespace(cls: type) -> tuple[str, ...]:
    """Infer init namespace from a given class."""
    init_signature = signature(cls.__init__)  # type: ignore
    namespace = tuple(init_signature.parameters.keys())
    return namespace[1:]  # Note: disregard `self`


def _infer_settings_namespace(cls: type) -> tuple[str, ...]:
    """Infer strategy settings namespace from a given class."""
    if cls.__init__ is object.__init__:  # type: ignore
        return ()  # Note: to avoid `*(*kw)args` from `object.__init__`
    init_namespace: tuple[str, ...] = _infer_init_namespace(cls)
    return tuple(name for name in init_namespace if not name.startswith("_"))


def _closest_common_ancestor(*args) -> type:
    """Retrieve closest common ancestor class from the input classes' MROs.

    Note: metaclasses are not supported, ``object`` is always shared.
    """
    cls_list = map(lambda obj: obj if isinstance(obj, type) else type(obj), args)
    mros = [cls.mro() for cls in cls_list]
    base = min(mros, key=len)
    mros.remove(base)
    for cls in base:
        if all(cls in mro for mro in mros):
            return cls
    return None  # Note: safeguard, `object` always shared (never called)  # pragma: no cover


def _shared_strategy_ancestor(*strategies) -> type:
    """Return closest common strategy ancestor or None."""
    shared: type = _closest_common_ancestor(*strategies)
    if shared in (None, *_BASE_STRATEGY_CLASSES) or not issubclass(shared, _BaseStrategy):
        return None
    return shared


################################################################################
## BASE STRATEGY CLASSES
################################################################################
class _BaseStrategy:
    """Strategy base class.

    Note: this class is not meant to be instantiated or inherited directly.
    """

    # TODO: subscriptable classconstant types (e.g. `classconstant[str]`)
    NAME: classconstant  # str
    INIT_NAMESPACE: classconstant  # tuple[str]
    SETTINGS_NAMESPACE: classconstant  # tuple[str]

    def __init_subclass__(cls, *args, **kwargs) -> None:
        super().__init_subclass__(*args, **kwargs)
        cls._validate_init()
        cls.NAME = classconstant(cls.__name__)
        cls.INIT_NAMESPACE = classconstant(_infer_init_namespace(cls))
        cls.SETTINGS_NAMESPACE = classconstant(_infer_settings_namespace(cls))

    def __new__(cls, *args, **kwargs) -> _BaseStrategy:
        # Note: logic added to `__new__` to avoid making `super().__init__` mandatory
        self = super().__new__(cls)
        self._original_args = _pack_init_args(self, *args, **kwargs)  # type: ignore
        return self

    def __init__(self) -> None:
        pass  # Note: dummy, to avoid `*(*kw)args` from `object.__init__`

    def __getattr__(self, name: str) -> Any:
        try:
            return super().__getattr__(name)  # type: ignore
        except (AttributeError, TypeError):
            original_args = getattr(self, "_original_args", {})
            attr = original_args.get(name, UNSET)
        if attr is UNSET:
            raise AttributeError(f"'{type(self)}' object has no attribute '{name}'")
        return deepcopy(attr)

    def __repr__(self) -> str:
        strategy_name: str = type(self).__name__
        settings_str = ", ".join(f"{key}={value!r}" for key, value in self.settings.items())
        return f"{strategy_name}({settings_str})" if settings_str else strategy_name

    def __eq__(self, other: object) -> bool:
        cls: type = _shared_strategy_ancestor(self, other)
        if cls is None or not self.is_facade(cls) or not other.is_facade(cls):  # type: ignore
            return False
        SHARED_NAMESPACE = cls.SETTINGS_NAMESPACE  # type: ignore # pylint: disable=invalid-name
        self_settings = {name: getattr(self, name, UNSET) for name in SHARED_NAMESPACE}
        other_settings = {name: getattr(other, name, UNSET) for name in SHARED_NAMESPACE}
        return self_settings == other_settings

    @property
    def settings(self) -> dict[str, Any]:
        """Strategy settings."""
        return {key: getattr(self, key) for key in self.SETTINGS_NAMESPACE}

    @property
    def init_args(self) -> dict[str, Any]:
        """Strategy init args."""
        return {key: getattr(self, key) for key in self.INIT_NAMESPACE}

    @property
    def original_args(self) -> dict[str, Any]:
        """A copy of original args."""
        original_args = getattr(self, "_original_args", {})
        return deepcopy(original_args)

    @classmethod
    def is_facade(cls, ancestor: Any) -> bool:
        """Checks if the strategy class is a facade subclass of the given ancestor."""
        if not isinstance(ancestor, type):
            ancestor = type(ancestor)
        if not (issubclass(cls, ancestor) and issubclass(ancestor, _BaseStrategy)):
            return False  # Note: subclassing is transitive (A -> B -> C then A -> C)
        filtered_mro = list(filter(lambda cls: issubclass(cls, ancestor), cls.mro()))
        for facade, parent in zip(filtered_mro, filtered_mro[1:]):
            if set(facade.SETTINGS_NAMESPACE) - set(parent.SETTINGS_NAMESPACE):  # type: ignore
                return False
        return True

    def replicate(self, **kwargs) -> _BaseStrategy:
        """Build a replica of the current strategy altering the specified settings."""
        init_args = deepcopy(self.init_args)
        init_args.update(kwargs)
        return type(self)(**init_args)

    @classmethod
    def _validate_init(cls):
        """Validate init method.

        Raises:
            TypeError: if POSITIONAL_ONLY, VAR_POSITIONAL, or VAR_KEYWORD arguments.
        """
        init_signature = signature(cls.__init__)
        parameters = init_signature.parameters.values()
        kinds = tuple(p.kind for p in parameters)[1:]  # Note: disregard `self`
        disallowed = {Parameter.POSITIONAL_ONLY, Parameter.VAR_POSITIONAL, Parameter.VAR_KEYWORD}
        for kind in kinds:
            if kind in disallowed:
                raise TypeError(
                    f"Invalid __init__ method in strategy class <{cls.__name__}>: "
                    f"{kind} arguments disallowed."
                )


# TODO
class _BaseStrategyABC(_BaseStrategy, ABC):
    """Strategy abstract base class.

    Note: this class is not meant to be instantiated or inherited directly.
    """


# TODO
class _FrozenStrategy(_BaseStrategy):
    """Frozen (immutable) strategy base class.

    Note: this class is not meant to be instantiated or inherited directly.
    """


# TODO
class _FrozenStrategyABC(_FrozenStrategy, ABC):
    """Frozen (immutable) strategy abstract base class.

    Note: this class is not meant to be instantiated or inherited directly.
    """


_BASE_STRATEGY_CLASSES: set[type] = {
    _BaseStrategy,
    _BaseStrategyABC,
    _FrozenStrategy,
    _FrozenStrategyABC,
}



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""ZNE functionality for :method:`qiskit.primitives.BaseEstimator.__call__` method."""
from __future__ import annotations

from collections.abc import Callable, Sequence
from functools import wraps

from numpy import ndarray
from qiskit.circuit import QuantumCircuit
from qiskit.primitives import EstimatorResult
from qiskit.quantum_info.operators import SparsePauliOp
from qiskit.utils.deprecation import deprecate_arguments

Sequence.register(ndarray)


################################################################################
## DECORATOR
################################################################################
def zne_call(call: Callable) -> Callable:
    """Add ZNE functionality to :method:`qiskit.primitives.BaseEstimator.__call__`."""

    if not callable(call):
        raise TypeError("Invalid `call` argument, expected callable.")

    @wraps(call)
    @deprecate_arguments({"circuit_indices": "circuits", "observable_indices": "observables"})
    def _zne_call(
        self,
        circuits: Sequence[int | QuantumCircuit],
        observables: Sequence[int | SparsePauliOp],
        parameter_values: Sequence[Sequence[float]] | None = None,
        **run_options,
    ) -> EstimatorResult:
        raise TypeError(
            "The BaseEstimator.__call__ method is deprecated as of Qiskit Terra 0.22.0. "
            "Use the 'run' method instead.",
        )

    return _zne_call



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Noise amplification library."""

from .folding_amplifier import (
    CxAmplifier,
    GlobalFoldingAmplifier,
    LocalFoldingAmplifier,
    MultiQubitAmplifier,
    TwoQubitAmplifier,
)
from .noise_amplifier import CircuitNoiseAmplifier, DAGNoiseAmplifier, NoiseAmplifier

NOISE_AMPLIFIER_LIBRARY = {
    cls.__name__: cls
    for cls in (
        GlobalFoldingAmplifier,
        LocalFoldingAmplifier,
        CxAmplifier,
        TwoQubitAmplifier,
        MultiQubitAmplifier,
    )
}

__all__ = [
    "NoiseAmplifier",
    "CircuitNoiseAmplifier",
    "DAGNoiseAmplifier",
    *NOISE_AMPLIFIER_LIBRARY.keys(),
]



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from unittest.mock import Mock

from numpy import array
from pytest import fixture, mark, raises
from qiskit.primitives import EstimatorResult

from zne.meta.job import ZNEJob
from zne.zne_strategy import ZNEStrategy


################################################################################
## FIXTURES
################################################################################
@fixture(scope="module")
def base_job():
    return "base_job"


@fixture(scope="module")
def zne_strategy():
    return "zne_strategy"


@fixture(scope="module")
def target_num_experiments():
    return "target_num_experiments"


################################################################################
## TESTS
################################################################################
def test_init(base_job, zne_strategy, target_num_experiments):
    job = ZNEJob(base_job, zne_strategy, target_num_experiments)
    assert job.base_job == base_job
    assert job.zne_strategy == zne_strategy
    assert job.target_num_experiments == target_num_experiments


def test_porperties(base_job, zne_strategy, target_num_experiments):
    job = ZNEJob(base_job, zne_strategy, target_num_experiments)
    assert job.base_job is job._base_job
    assert job.zne_strategy is job._zne_strategy
    assert job.target_num_experiments is job._target_num_experiments


@mark.parametrize(
    "target_num_experiments, num_noise_factors",
    [
        (1, 1),
        (1, 2),
        (2, 1),
        (2, 2),
    ],
)
def test_result(target_num_experiments, num_noise_factors):
    # Results
    values = [0] * target_num_experiments
    noisy_values = values * num_noise_factors
    result = EstimatorResult(values=array(values), metadata=[{}] * len(values))
    noisy_result = EstimatorResult(values=array(noisy_values), metadata=[{}] * len(noisy_values))

    # Jobs
    base_job = Mock()
    base_job.result = Mock(return_value=noisy_result)
    zne_strategy = ZNEStrategy(noise_factors=range(1, num_noise_factors + 1))
    zne_strategy.mitigate_noisy_result = Mock(return_value=result)
    job = ZNEJob(base_job, zne_strategy, target_num_experiments)

    # Testing
    if zne_strategy.performs_zne:
        assert job.result() is result
    else:
        assert job.result() is noisy_result


@mark.parametrize(
    "target_num_experiments, num_noise_factors",
    [
        (1, 1),
        (1, 2),
        (2, 1),
        (2, 2),
    ],
)
def test_result_runtime_error(target_num_experiments, num_noise_factors):
    # Results
    values = [0] * (target_num_experiments + 1)  # Introducing discrepancy
    noisy_values = values * num_noise_factors
    result = EstimatorResult(values=array(values), metadata=[{}] * len(values))
    noisy_result = EstimatorResult(values=array(noisy_values), metadata=[{}] * len(noisy_values))

    # Jobs
    base_job = Mock()
    base_job.result = Mock(return_value=noisy_result)
    zne_strategy = ZNEStrategy(noise_factors=range(1, num_noise_factors + 1))
    zne_strategy.mitigate_noisy_result = Mock(return_value=result)
    job = ZNEJob(base_job, zne_strategy, target_num_experiments)

    # Testing
    with raises(RuntimeError):
        job.result()


################################################################################
## NESTED JOB DELEGATION
################################################################################
@mark.parametrize(
    "attr",
    "When things are in danger someone has to give them up so that others may keep them".split(),
)
def test_getattribute(zne_strategy, target_num_experiments, attr):
    base_job = Mock()
    setattr(base_job, attr, f"value of {attr}")
    job = ZNEJob(base_job, zne_strategy, target_num_experiments)
    assert getattr(job, attr) is getattr(base_job, attr)


@mark.parametrize("attr", ZNEJob._ZNE_ATTRIBUTES)
def test_getattribute_zne(base_job, zne_strategy, target_num_experiments, attr):
    job = ZNEJob(base_job, zne_strategy, target_num_experiments)
    assert getattr(job, attr) is not getattr(base_job, attr, f"base_{attr}")


@mark.parametrize(
    "attr",
    "All we have to decide is what to do with the time that is given to us".split(),
)
def test_setattr(zne_strategy, target_num_experiments, attr):
    base_job = Mock()
    job = ZNEJob(base_job, zne_strategy, target_num_experiments)
    setattr(job, attr, f"value of {attr}")
    assert getattr(job, attr) is getattr(base_job, attr)


@mark.parametrize(
    "attr",
    [
        attr
        for attr in ZNEJob._ZNE_ATTRIBUTES
        if not isinstance(getattr(ZNEJob, attr, None), property)
    ],
)
def test_setattr_zne(base_job, zne_strategy, target_num_experiments, attr):
    job = ZNEJob(base_job, zne_strategy, target_num_experiments)
    setattr(job, attr, f"value of {attr}")
    assert getattr(job, attr) is object.__getattribute__(job, attr)
    assert getattr(job, attr) is not getattr(base_job, attr, f"base_{attr}")


@mark.parametrize(
    "attr",
    "And some things that should not have been forgotten were lost".split(),
)
def test_delattr(zne_strategy, target_num_experiments, attr):
    base_job = Mock()
    setattr(base_job, attr, f"value of {attr}")
    job = ZNEJob(base_job, zne_strategy, target_num_experiments)
    delattr(job, attr)
    NOT_FOUND = Mock()
    assert getattr(base_job, attr, NOT_FOUND) is NOT_FOUND


@mark.parametrize(
    "attr",
    [
        attr
        for attr in ZNEJob._ZNE_ATTRIBUTES
        if not isinstance(getattr(ZNEJob, attr, None), property)
        and not isinstance(getattr(ZNEJob, attr, None), type(lambda: ...))
    ],
)
def test_delattr_zne(zne_strategy, target_num_experiments, attr):
    base_job = Mock()
    base_attr = Mock()
    setattr(base_job, attr, base_attr)
    job = ZNEJob(base_job, zne_strategy, target_num_experiments)
    delattr(job, attr)
    NOT_FOUND = Mock()
    assert getattr(job, attr, NOT_FOUND) is NOT_FOUND
    assert getattr(base_job, attr, NOT_FOUND) is base_attr


################################################################################
## DUMMY ABSTRACT METHOD IMPLEMENTATION
################################################################################
def test_submit(zne_strategy, target_num_experiments):
    base_job = Mock()
    SUBMITTED = Mock()
    base_job.submit = Mock(return_value=SUBMITTED)
    job = ZNEJob(base_job, zne_strategy, target_num_experiments)
    submit = object.__getattribute__(job, "submit")  # Bypass nested delegation logic
    assert submit() is SUBMITTED


def test_status(zne_strategy, target_num_experiments):
    base_job = Mock()
    STATUS = Mock()
    base_job.status = Mock(return_value=STATUS)
    job = ZNEJob(base_job, zne_strategy, target_num_experiments)
    status = object.__getattribute__(job, "status")  # Bypass nested delegation logic
    assert status() is STATUS



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Noise amplification strategy via gate/inverse repetition."""

from __future__ import annotations

from collections.abc import Iterable, Sequence

from qiskit.circuit import CircuitInstruction, QuantumCircuit

from ...utils import STANDARD_GATES
from ...utils.docstrings import insert_into_docstring
from .folding_amplifier import FoldingAmplifier


################################################################################
## GENERAL
################################################################################
class LocalFoldingAmplifier(FoldingAmplifier):
    """
    Replaces gates in the circuit by said gate and its inverse alternatingly as many times as
    indicated by noise_factor. The gates that should be folded can be specified by
    ``gates_to_fold``. By default, all gates of the circuit are folded.

    If ``noise_factor`` is not an odd integer, a subset of gates of the full circule are sub-folded
    such that the resultant noise scaling :math:`\\lambda` equals :math:`\\lambda=(2n+1)+2s/d`,
    where :math:`d` are the number of layers of the original circuit, :math:`n` is the number of
    gate foldings, and :math:`s` is the number of gates that are sub-folded [1].

    The gates of the full circuit that are sub-folded can be the first :math:`s` gates, the last
    :math:`s` gates or a random selection of :math:`s` gates. This is specified by
    ``sub_folding_option``.

    Note that all instance objects and their attributes are immutable by construction. If another
    strategy with different options needs to be utilized, a new instance of the class should be
    created.

    References:
        [1] T. Giurgica-Tiron et al. (2020).
            Digital zero noise extrapolation for quantum error mitigation.
            `<https://ieeexplore.ieee.org/document/9259940>`
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        gates_to_fold: Sequence[int | str] | str | int | None = None,
        sub_folding_option: str = "from_first",
        barriers: bool = True,
        random_seed: int | None = None,
        noise_factor_relative_tolerance: float = 1e-2,
        warn_user: bool = True,
    ) -> None:
        self.warn_user: bool = warn_user
        self._set_gates_to_fold(gates_to_fold)  # TODO move to super eventually
        super().__init__(
            sub_folding_option=sub_folding_option,
            barriers=barriers,
            random_seed=random_seed,
            noise_factor_relative_tolerance=noise_factor_relative_tolerance,
            warn_user=warn_user,
        )

    # TODO: remove `insert_into_docstring`
    __init__.__doc__ = insert_into_docstring(
        FoldingAmplifier.__init__.__doc__,
        [
            (
                "Args",
                "gates_to_fold: Specifies which gates of the full circuit are folded. It \n"
                "can be an int (referring to all n-qubit gates), a str with the name of a \n"
                "gate, or a sequence of int/str. In the latter case the union of all specified \n"
                "gates will be folded.",
            ),
        ],
    )

    ################################################################################
    ## PROPERTIES
    ################################################################################
    @property
    def gates_to_fold(self) -> frozenset[int | str] | None:
        """Set of gates that are folded.

        Specifies which gates of the full circuit are folded. By default (i.e. ``None``) all gates
        are folded). Otherwise, ``gates_to_fold`` is a sequence of strings and/or integers where
        each string indicates the name of a gate to be folded and an integer n indicates that all
        n-qubit gates should be folded. For example, (2, "h", "x") would specify that the union,
        i.e., all two-qubit gates, the hadamard gate, and the x-gate will be noise amplified.
        """
        return self._gates_to_fold

    def _set_gates_to_fold(self, gates_to_fold: Sequence[int | str] | str | int | None) -> None:
        self._gates_to_fold: frozenset[int | str] | None = self._parse_gates_to_fold(gates_to_fold)

    def _parse_gates_to_fold(
        self, gates_to_fold: Sequence[int | str] | str | int | None
    ) -> frozenset[int | str] | None:
        """Parse to immutable set.

        Args:
            gates_to_fold: Set of gates that are folded.

        Raises:
            TypeError: If gates_to_fold is neither None, int, str, or an iterable.
        """
        if gates_to_fold is None:
            return gates_to_fold
        if isinstance(gates_to_fold, (int, str)):
            gates_to_fold_set: frozenset[int | str] = frozenset((gates_to_fold,))
        elif isinstance(gates_to_fold, Iterable):
            gates_to_fold_set = frozenset(gates_to_fold)
        else:
            raise TypeError(
                f"Expected int, str, or Iterable, received {type(gates_to_fold)} instead."
            )
        self._validate_gates_to_fold(gates_to_fold_set)
        return gates_to_fold_set

    def _validate_gates_to_fold(self, gates_to_fold: frozenset[int | str]) -> None:
        """Validates gates_to_fold set.

        Args:
            gates_to_fold: Set of gates that are folded.

        Raises:
            TypeError: If gates_to_fold contains objects that are neither int nor str.
            ValueError: If gates_to_fold contains integers smaller than one.
        """
        if not all(isinstance(g, (int, str)) for g in gates_to_fold):
            raise TypeError(
                f"Expected Iterable[str | int], received {type(gates_to_fold)} instead."
            )
        for gate in gates_to_fold:
            if isinstance(gate, int) and gate < 1:
                raise ValueError(f"gates_to_fold contains {gate} which is smaller than one.")
            if isinstance(gate, str) and gate not in STANDARD_GATES:
                self.warn(f"gates_to_fold contains {gate} which is not a standard gate name.")

    ################################################################################
    ## INTERFACE
    ################################################################################
    def amplify_circuit_noise(self, circuit: QuantumCircuit, noise_factor: float) -> QuantumCircuit:
        self._validate_noise_factor(noise_factor)
        foldings: list[int] = self._build_foldings_per_gate(circuit, noise_factor)
        noisy_circuit = circuit.copy_empty_like()
        for operation, num_foldings in zip(circuit, foldings):
            if num_foldings == 0:
                noisy_circuit.append(*operation)
            else:
                noisy_circuit = self._append_folded(noisy_circuit, operation, num_foldings)
        return noisy_circuit

    def _build_foldings_per_gate(self, circuit: QuantumCircuit, noise_factor: float) -> list[int]:
        """Returns number of foldings for each gate in the circuit.

        Args:
            circuit: The original circuit.
            noise_factor: The (float) noise amplification factor by which to amplify the overall
            circuit noise.

        Returns:
            List of (int) number of foldings for each gate in the circuit.
        """
        gates_to_fold_mask = [self._check_gate_folds(operation) for operation in circuit.data]
        num_gates_to_fold: int = gates_to_fold_mask.count(True)
        foldings = self._build_foldings(noise_factor, num_gates_to_fold)
        return [foldings.pop(0) if m else 0 for m in gates_to_fold_mask]

    def _check_gate_folds(self, operation: CircuitInstruction) -> bool:
        """Checks whether circuit operation should be folded.

        Args:
            operation: A single circuit instruction of the original circuit.

        Returns:
            True if instruction should be folded, False otherwise.
        """
        instruction, qargs, _ = operation
        if instruction.name in {"barrier", "measure"}:
            return False
        num_qubits = len(qargs)
        return (
            (self._gates_to_fold is None)
            or (num_qubits in self._gates_to_fold)
            or (instruction.name in self._gates_to_fold)
        )

    def _build_foldings(self, noise_factor: float, num_gates_to_fold: int) -> list[int]:
        """Returns number of foldings for each gate to fold.

        Args:
            noise_factor: The (float) noise amplification factor by which to amplify the overall
            circuit noise.
            num_gates_to_fold: The total number of gates to fold.

        Returns:
            List of (int | None) number of foldings for each gate to fold.
        """
        num_full_foldings, num_sub_foldings = self._compute_folding_nums(
            noise_factor, num_gates_to_fold
        )
        sub_foldings: list[int] = self._build_sub_foldings(num_sub_foldings, num_gates_to_fold)
        return [num_full_foldings + sf for sf in sub_foldings]

    def _build_sub_foldings(self, num_sub_foldings: int, num_gates_to_fold: int) -> list[int]:
        """Returns sub-foldings for each gate to fold.

        Args:
            num_sub_foldings: Number of gates to be sub-folded.
            num_gates_to_fold: The number of instructions to fold.

        Returns:
            List of number of sub-foldings for each folded gate.
        """
        if num_sub_foldings == 0:
            return [0] * num_gates_to_fold
        if self.sub_folding_option == "from_first":
            return [1] * num_sub_foldings + [0] * (num_gates_to_fold - num_sub_foldings)
        if self.sub_folding_option == "from_last":
            return [0] * (num_gates_to_fold - num_sub_foldings) + [1] * num_sub_foldings
        idxs = self._rng.choice(num_gates_to_fold, size=num_sub_foldings, replace=False)
        return [1 if i in idxs else 0 for i in range(num_gates_to_fold)]

    def _append_folded(
        self, noisy_circuit: QuantumCircuit, operation: CircuitInstruction, num_foldings: int
    ) -> QuantumCircuit:
        """Folds circuit operation.

        Args:
            noisy_circuit: The noise amplified circuit to which the gate foldings are added.
            operation: A single operation of the original circuit to be folded.
            noise_factor: The corresponding noise factor.

        Returns:
            The updated noisy circuit.
        """
        # TODO: Create FoldableCircuit class extending QuantumCircuit
        # TODO: CircuitInstruction.inverse()
        self._validate_num_foldings(num_foldings)
        instruction, qargs, cargs = operation
        noisy_circuit = self._apply_barrier(noisy_circuit, qargs)
        noisy_circuit.append(instruction, qargs, cargs)
        if num_foldings > 0:
            noisy_circuit = self._apply_barrier(noisy_circuit, qargs)
            noisy_circuit.append(instruction.inverse(), qargs, cargs)
            return self._append_folded(noisy_circuit, operation, num_foldings - 1)
        noisy_circuit = self._apply_barrier(noisy_circuit, qargs)
        return noisy_circuit

    @staticmethod
    def _validate_num_foldings(num_foldings: int) -> None:
        """Validate num_foldings.

        Args:
            num_foldings: number of foldings.

        Raises:
            TypeError: If num_foldings is not int.
            ValueError: If num_foldings is negative.
        """
        if not isinstance(num_foldings, int):
            raise TypeError(f"Expected int, received {type(num_foldings)} instead.")
        if num_foldings < 0:
            raise ValueError("Number of foldings must be greater than or equal to zero.")


################################################################################
## FACADES
################################################################################
class CxAmplifier(LocalFoldingAmplifier):
    """
    Digital noise amplifier acting on CX gates exclusively.

    Replaces each CX gate locally with as many CX gates as indicated by the noise_factor.
    """

    def __init__(  # pylint: disable=too-many-arguments,duplicate-code
        self,
        sub_folding_option: str = "from_first",
        barriers: bool = True,
        random_seed: int | None = None,
        noise_factor_relative_tolerance: float = 1e-2,
        warn_user: bool = True,
    ) -> None:
        super().__init__(
            gates_to_fold="cx",
            sub_folding_option=sub_folding_option,
            barriers=barriers,
            random_seed=random_seed,
            noise_factor_relative_tolerance=noise_factor_relative_tolerance,
            warn_user=warn_user,
        )


class TwoQubitAmplifier(LocalFoldingAmplifier):
    """
    Digital noise amplifier acting on two-qubit gates exclusively.

    Replaces each two-qubit gate locally with as many gates as indicated by the noise_factor.
    """

    def __init__(  # pylint: disable=too-many-arguments,duplicate-code
        self,
        sub_folding_option: str = "from_first",
        barriers: bool = True,
        random_seed: int | None = None,
        noise_factor_relative_tolerance: float = 1e-2,
        warn_user: bool = True,
    ) -> None:
        super().__init__(
            gates_to_fold=2,
            sub_folding_option=sub_folding_option,
            barriers=barriers,
            random_seed=random_seed,
            noise_factor_relative_tolerance=noise_factor_relative_tolerance,
            warn_user=warn_user,
        )


class MultiQubitAmplifier(LocalFoldingAmplifier):
    """
    Digital noise amplifier acting on multi-qubit gates exclusively.

    Replaces each multi-qubit gate locally with as many gates as indicated by the noise_factor.
    """

    def __init__(  # pylint: disable=too-many-arguments,duplicate-code
        self,
        sub_folding_option: str = "from_first",
        barriers: bool = True,
        random_seed: int | None = None,
        noise_factor_relative_tolerance: float = 1e-2,
        warn_user: bool = True,
    ) -> None:
        super().__init__(
            gates_to_fold=None,
            sub_folding_option=sub_folding_option,
            barriers=barriers,
            random_seed=random_seed,
            noise_factor_relative_tolerance=noise_factor_relative_tolerance,
            warn_user=warn_user,
        )

    def _check_gate_folds(self, operation: CircuitInstruction) -> bool:
        instruction, qargs, _ = operation
        if instruction.name in {"barrier", "measure"}:
            return False
        num_qubits = len(qargs)
        return num_qubits > 1



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Polynomial extrapolator."""

from __future__ import annotations

from collections import namedtuple

from numpy import array, ndarray, sqrt, zeros
from scipy.optimize import curve_fit

from .extrapolator import OLSExtrapolator, ReckoningResult

################################################################################
## GENERAL
################################################################################
_RegressionData = namedtuple("_RegressionData", ("x_data", "y_data", "sigma_x", "sigma_y"))


class PolynomialExtrapolator(OLSExtrapolator):
    """Polynomial ordinary-least-squares (OLS) extrapolator.

    Args:
        degree: The degree of the polynomial regression curve.
    """

    def __init__(self, degree: int = 1):  # pylint: disable=super-init-not-called
        self._set_degree(degree)

    ################################################################################
    ## PROPERTIES
    ################################################################################
    @property
    def degree(self) -> int:
        """The degree of the regression polynomial."""
        return self._degree

    def _set_degree(self, degree: int) -> None:
        """Degree setter."""
        degree = int(degree)
        if degree < 1:
            raise ValueError("Polynomial degree must be at least 1.")
        self._degree: int = degree

    ################################################################################
    ## IMPLEMENTATION
    ################################################################################
    @property
    def min_points(self) -> int:
        return self.degree + 1

    # pylint: disable=duplicate-code
    def _extrapolate_zero(
        self,
        x_data: tuple[float, ...],
        y_data: tuple[float, ...],
        sigma_x: tuple[float, ...],
        sigma_y: tuple[float, ...],
    ) -> ReckoningResult:
        # TODO: if curve fit fails (e.g. p-value test) warn and return closest to zero
        regression_data = _RegressionData(x_data, y_data, sigma_x, sigma_y)
        return self._infer(0, regression_data)

    def _model(self, x, *coefficients) -> ndarray:  # pylint: disable=invalid-name
        """Polynomial regression model for curve fitting."""
        x = array(x)
        return sum(c * (x**i) for i, c in enumerate(coefficients))

    ################################################################################
    ## AUXILIARY
    ################################################################################
    def _infer(self, target: float, regression_data: _RegressionData) -> ReckoningResult:
        """Fit regression model from data and infer evaluation for target value.

        Args:
            target: The target X value to infer a Y value for.
            regression_data: A four-tuple of tuples representing X-data, Y-data,
                and corresponding std errors for the X and Y data respectively.

        Returns:
            Reckoning result holding the inferred value, std error, and metadata about
            the curve fit procedure.
        """
        coefficients, covariance_matrix = curve_fit(
            self._model,
            regression_data.x_data,
            regression_data.y_data,
            sigma=self._compute_sigma(regression_data.y_data, regression_data.sigma_y),
            absolute_sigma=True,
            p0=zeros(self.degree + 1),  # Note: Initial point determines number of d.o.f.
        )
        target_powers = array([target**p for p in range(self.degree + 1)])
        value = target_powers @ coefficients  # Note: == self._model(target, *coefficients)
        variance = target_powers @ covariance_matrix @ target_powers
        std_error = sqrt(variance)
        metadata = self._build_metadata(
            array(regression_data.x_data),
            array(regression_data.y_data),
            coefficients,
            covariance_matrix,
        )
        return ReckoningResult(value.tolist(), std_error.tolist(), metadata)


################################################################################
## FACADES
################################################################################
class LinearExtrapolator(PolynomialExtrapolator):
    """Linear ordinary-least-squares (OLS) extrapolator."""

    def __init__(self):
        super().__init__(degree=1)


class QuadraticExtrapolator(PolynomialExtrapolator):
    """Quadratic ordinary-least-squares (OLS) extrapolator."""

    def __init__(self):
        super().__init__(degree=2)


class CubicExtrapolator(PolynomialExtrapolator):
    """Cubic ordinary-least-squares (OLS) extrapolator."""

    def __init__(self):
        super().__init__(degree=3)


class QuarticExtrapolator(PolynomialExtrapolator):
    """Quartic ordinary-least-squares (OLS) extrapolator."""

    def __init__(self):
        super().__init__(degree=4)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022-2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from pytest import mark, raises

from zne.utils.classconstant import classconstant


class TestClassConstant:
    """Test `classconstant` descriptor."""

    @mark.parametrize("value", [0, 1, "pi"])
    def test_get(self, value):
        """Test get."""
        cls = type("cls", (), {"CONSTANT": classconstant(value)})
        assert cls.CONSTANT == value
        assert cls().CONSTANT == value

    def test_set(self):
        """Test set."""
        cls = type("cls", (), {"CONSTANT": classconstant(0)})
        with raises(AttributeError):
            cls().CONSTANT = "value"

    def test_delete(self):
        """Test delete."""
        cls = type("cls", (), {"CONSTANT": classconstant(0)})
        with raises(AttributeError):
            del cls().CONSTANT
